MINIMUM_DATASET_SHACL = """
@prefix sh:    <http://www.w3.org/ns/shacl#> .
@prefix dcat:  <http://www.w3.org/ns/dcat#> .
@prefix spdx:  <http://spdx.org/rdf/terms#> .
@prefix xsd:   <http://www.w3.org/2001/XMLSchema#> .
@prefix ex:    <http://example.org/ns#> .

ex:DatasetShape
  a sh:NodeShape ;
  sh:targetClass dcat:Dataset ;
  sh:property [
    sh:path dcat:distribution ;
    sh:minCount 1 ;
    sh:node ex:DistributionShape ;
    sh:message "A dcat:Dataset must have at least one dcat:distribution, each conforming to DistributionShape." ;
  ] .


ex:DistributionShape
  a sh:NodeShape ;
  sh:targetClass dcat:Distribution ;

  # Require a downloadURL (IRI)
  sh:property [
    sh:path dcat:downloadURL ;
    sh:minCount 1 ;
    sh:nodeKind sh:IRI ;
    sh:message "A dcat:Distribution must have at least one dcat:downloadURL (IRI)." ;
  ] ;

  # Require a mediaType (IRI)
  sh:property [
    sh:path dcat:mediaType ;
    sh:minCount 1 ;
    sh:nodeKind sh:IRI ;
    sh:message "A dcat:Distribution must have at least one dcat:mediaType (IRI)." ;
  ] .

  # # Require at least one checksum (without structural validation)
  # sh:property [
  #   sh:path spdx:checksum ;
  #   sh:minCount 1 ;
  #   sh:message "A dcat:Distribution must have at least one spdx:checksum." ;
  # ] .
"""


import enum
import os
import pathlib
import shutil
from collections import defaultdict
from dataclasses import dataclass
from typing import List, Union, Optional, Type

import dotenv
import rdflib
import requests
import requests.exceptions
import tqdm
from h5rdmtoolbox import catalog as h5cat
from h5rdmtoolbox.catalog import InMemoryRDFStore, GraphDB, HDF5FileStore
from h5rdmtoolbox.catalog.profiles import IS_VALID_CATALOG_SHACL
from h5rdmtoolbox.repository.zenodo import ZenodoRecord
from ontolutils import Thing
from ontolutils.ex import dcat, qudt, sosa, ssn
from ontolutils.ex.sis import StandardMU
from ontolutils.ex.sosa import Observation
from rdflib.namespace import split_uri
from ssnolib import StandardName
from ssnolib.m4i import NumericalVariable

from opencefadb import logger
from opencefadb import sparql_templates
from opencefadb._core._database_initialization import DownloadStatus, database_initialization
# from opencefadb.sparql_templates.sparql import get_properties, find_dataset_value_in_same_group_by_other_standard_name, \
#     find_datasets_by_standard_name_and_value_range
from opencefadb.utils import download_file, remove_none
from .models.wikidata import FAN_OPERATING_POINT
from .utils import opencefa_print

SANDBOX_BASE_URL = "https://sandbox.zenodo.org/api/records/428370"
PRODUCTION_BASE_URL = "https://zenodo.org/api/records/18349358"
__this_dir__ = pathlib.Path(__file__).parent

_db_instance = None

unit_entities = {
    "http://qudt.org/vocab/unit/PA": qudt.Unit(
        id="http://qudt.org/vocab/unit/PA",
        name="Pascal",
        symbol="Pa",
        conversionMultiplier=1.0,
    )
}


def parse_to_entity(df, uri, entity: Type[Thing]):
    # check if columns are "property" and "value":
    if "property" not in df.columns or "value" not in df.columns:
        raise ValueError("DataFrame must contain 'property' and 'value' columns.")
    _grouped_dict = df.groupby("property")["value"].apply(list).to_dict()
    # extract the value from the uri, which is in the key
    _data_dict = {
        "id": uri
    }
    for k, v in _grouped_dict.items():
        _, key = split_uri(k)
        if len(v) == 1:
            _data_dict[key] = v[0]
        else:
            _data_dict[key] = v
    return entity.model_validate(_data_dict)


def get_and_unpack_property_value_query(uri: str, entity: Type[Thing], metadata_store: h5cat.RDFStore) -> Thing:
    """Evaluates a SPARQL query against the metadata store to retrieve the unit entity for the given unit URI."""
    sparql_query = sparql_templates.get_properties(uri)
    res = sparql_query.execute(metadata_store)
    if not res:
        return entity(id=uri)
    df = res.data
    return parse_to_entity(df, uri, entity)


def get_unit_entity(unit_uri: str, metadata_store: h5cat.RDFStore) -> Optional[qudt.Unit]:
    """Evaluates a SPARQL query against the metadata store to retrieve the unit entity for the given unit URI."""
    return get_and_unpack_property_value_query(unit_uri, qudt.Unit, metadata_store)


def get_standard_name_entity(standard_name_uri: str, metadata_store: h5cat.RDFStore) -> Thing:
    return get_and_unpack_property_value_query(standard_name_uri, StandardName, metadata_store)


@dataclass
class DistributionMetadata:
    download_url: str
    media_type: Optional[str] = None
    size: Optional[str] = None
    checksum: Optional[str] = None
    checksum_algorithm: Optional[str] = None


class MediaType(enum.Enum):
    JSON_LD = "application/ld+json"
    TURTLE = "text/turtle"
    IGES = "model/iges"
    IGS = "igs"
    CSV = "text/csv"
    TXT = "text/plain"
    XML = "application/rdf+xml"
    XML2 = "application/xml"
    HDF5 = "application/x-hdf5"

    @classmethod
    def parse(cls, media_type: str):
        media_type = str(media_type)
        if media_type is None:
            return None
        if media_type.startswith("https://"):
            media_type = str(media_type).rsplit('media-types/', 1)[-1]
        elif media_type.startswith("http://"):
            media_type = str(media_type).rsplit('media-types/', 1)[-1]
        try:
            return cls(media_type)
        except ValueError:
            return None

    def get_suffix(self):
        if self == MediaType.JSON_LD:
            return ".jsonld"
        elif self == MediaType.HDF5:
            return ".hdf5"
        elif self == MediaType.TURTLE:
            return ".ttl"
        elif self in (MediaType.IGES, MediaType.IGS):
            return ".igs"
        elif self == MediaType.CSV:
            return ".csv"
        elif self == MediaType.TXT:
            return ".txt"
        elif self == MediaType.XML:
            return ".xml"
        elif self == MediaType.XML2:
            return ".xml"
        else:
            return ""


def _get_download_urls_of_metadata_distributions_of_publisher(
        publisher: str,
        doi: str
):
    publisher = str(publisher)
    if publisher.lower() != "zenodo":
        raise ValueError(f"Unsupported publisher: {publisher}")
    return _get_download_urls_of_metadata_distributions_of_zenodo_record(doi)


def _get_download_urls_of_metadata_distributions_of_zenodo_record(doi: str) -> List[DistributionMetadata]:
    _doi = str(doi)
    sandbox = "10.5072/zenodo" in _doi
    record_id = _doi.rsplit('zenodo.', 1)[-1]
    z = ZenodoRecord(source=int(record_id), sandbox=sandbox)
    return [DistributionMetadata(
        download_url=file.download_url,
        size=file.size,
        media_type=file.media_type,
        checksum=file.checksum,
        checksum_algorithm=file.checksum_algorithm or "md5"
    ) for filename, file in
        z.files.items() if
        filename.endswith('.ttl') or filename.endswith('.jsonld')]


# def _get_metadata_datasets(
#         config_filename: Union[str, pathlib.Path],
#         config_dir: Union[str, pathlib.Path]
# ) -> rdflib.Graph:
#     """Parses the configuration file and returns the graph with metadata datasets.
#     The configuration file should contain DCAT descriptions of datasets.
#
#     Parameters
#     ----------
#     config_filename : str or pathlib.Path
#         The path to the configuration file (e.g., config.ttl or config.jsonld).
#     config_dir : str or pathlib.Path
#         The directory where the configuration file will be copied to and parsed from.
#     """
#     config_filename = pathlib.Path(config_filename)
#     if not config_filename.exists():
#         raise FileNotFoundError(f"Configuration file not found: {config_filename}")
#     config_dir = pathlib.Path(config_dir)
#     if not config_dir.exists():
#         raise FileNotFoundError(f"Configuration directory not found: {config_dir}")
#     config_suffix = config_filename.suffix
#     db_dataset_config_filename = config_dir / f"db-dataset-config.{config_suffix}"
#     shutil.copy(config_filename, db_dataset_config_filename)
#
#     logger.debug(f"Parsing database dataset config '{db_dataset_config_filename.resolve().absolute()}'...")
#
#     if config_suffix == '.ttl':
#         fmt = "ttl"
#     elif config_suffix in ('.json', '.jsonld', '.json-ld'):
#         fmt = "json-ld"
#     else:
#         raise ValueError(f"Unsupported config file suffix: {config_suffix}")
#     g = rdflib.Graph()
#     g.parse(source=db_dataset_config_filename, format=fmt)
#     logger.debug("Successfully parsed database dataset config.")
#     return g


class OpenCeFaDB(h5cat.CatalogManager):

    def __init__(
            self,
            working_directory: Union[str, pathlib.Path],
            version: Union[str, pathlib.Path] = "latest",
            sandbox: bool = False
    ):
        """Initializes the OpenCeFaDB database.

        Parameters
        ----------
        working_directory : Union[str, pathlib.Path]
            The working directory where the database files will be stored.
        version : str, optional
            The version of the catalog to use. This can be a version string like "1.0.0", "latest", or a path to a local
            catalog file. Default is "latest".
        sandbox : bool, optional
            Whether to use the sandbox version of the catalog. Default is False.
        """
        if pathlib.Path(version).exists():
            opencefa_print("Using local catalog file as version.")
            catalog = dcat.Catalog.from_file(source=str(version))[0]
        else:
            catalog = self.__class__.download_catalog(
                version,
                sandbox=sandbox,
                target_directory=working_directory
            )
        working_directory = working_directory / f"opencefadb-{catalog.version}"
        working_directory.mkdir(parents=True, exist_ok=True)
        super().__init__(
            catalog=catalog,
            working_directory=working_directory
        )

    @classmethod
    def from_graphdb_setup(
            cls,
            *,
            working_directory: Union[str, pathlib.Path],
            version: Union[str, pathlib.Path] = "latest",
            sandbox: bool = False,
            endpoint: str = "http://localhost:7200",
            repository: str = "OpenCeFaDB",
            username: str = None,
            password: str = None,
            add_wikidata_store: bool = False
    ) -> "OpenCeFaDB":
        try:
            gdb = GraphDB(
                endpoint=endpoint,
                repository=repository,
                username=username,
                password=password
            )
            gdb.get_repository_info(repository)
        except requests.exceptions.ConnectionError as e:
            raise RuntimeError(f"GraphDB not available: {e}")
        gdb_exists = gdb.get_repository_info(repository)
        if not gdb_exists:
            raise RuntimeError(f"GraphDB repository '{repository}' does not exist.")
        working_directory = pathlib.Path(working_directory)
        db = cls(working_directory=working_directory, version=version, sandbox=sandbox)
        db.add_main_rdf_store(gdb)
        if add_wikidata_store:
            db.add_wikidata_store(augment_main_rdf_store=True)
        db.add_hdf_infile_index()
        hdf_store = HDF5FileStore(data_directory=db.hdf_directory)
        db.add_hdf_store(hdf_store)
        return db

    @classmethod
    def from_rdflib_setup(
            cls,
            *,
            working_directory: Union[str, pathlib.Path],
            version: Union[str, pathlib.Path] = "latest",
            sandbox: bool = False,
            add_wikidata_store: bool = False
    ) -> "OpenCeFaDB":
        working_directory = pathlib.Path(working_directory)
        working_directory.mkdir(parents=True, exist_ok=True)
        db = cls(working_directory=working_directory, version=version, sandbox=sandbox)
        opencefa_print(f"RDF files are/will be stored in the directory: {db.rdf_directory}")
        opencefa_print(f"HDF files are/will be stored in the directory: {db.hdf_directory}")

        InMemoryRDFStore.__populate_on_init__ = False
        local_rdf_store = InMemoryRDFStore(
            data_dir=db.rdf_directory,
            recursive_exploration=True,
            formats=["ttl"]
        )
        db.add_main_rdf_store(local_rdf_store)
        opencefa_print("Checking if metadata files need to be downloaded. If yes, this will take a moment...")
        db.download_metadata()
        opencefa_print("RDF data ready to use.")
        _local_graph_file = db.rdf_directory / ".graph.nt"
        if _local_graph_file.exists():
            g = rdflib.Graph().parse(_local_graph_file)
            local_rdf_store.graph += g
        else:
            local_rdf_store.populate(recursive=True)

        hdf_store = HDF5FileStore(data_directory=db.hdf_directory)
        db.add_hdf_store(hdf_store)

        if add_wikidata_store:
            db.add_wikidata_store(augment_main_rdf_store=True)

        db.add_hdf_infile_index()
        return db

    @classmethod
    def download_catalog(cls,
                         version: Optional[str] = None,
                         target_directory: Optional[Union[str, pathlib.Path]] = None,
                         sandbox: bool = False,
                         validate=True) -> dcat.Catalog:
        """Download the catalog (dcat:Catalog)"""
        catalog_filename = _download_catalog(version, target_directory, sandbox)
        catalog = dcat.Catalog.from_file(source=catalog_filename)[0]
        if validate:
            opencefa_print("Validating downloaded catalog against SHACL shapes...")
            catalog.validate(shacl_data=IS_VALID_CATALOG_SHACL, raise_on_error=True)
            opencefa_print("Catalog is valid.")
        return catalog

    @classmethod
    def initialize(
            cls,
            config_filename: Union[str, pathlib.Path],
            working_directory: Union[str, pathlib.Path] = None
    ) -> List[DownloadStatus]:
        if working_directory is None:
            working_directory = pathlib.Path.cwd()
        download_directory = pathlib.Path(working_directory) / "metadata"
        download_directory.mkdir(parents=True, exist_ok=True)
        opencefa_print(
            f"Copying the config file {config_filename} to the target directory {download_directory.resolve()} ..."
        )
        shutil.copy(
            config_filename,
            download_directory / pathlib.Path(config_filename).name
        )
        return database_initialization(
            config_filename=config_filename,
            download_directory=download_directory
        )

    @classmethod
    def get_config(cls, sandbox=False) -> pathlib.Path:
        if sandbox:
            return __this_dir__ / "data/opencefadb-config-sandbox.ttl"
        return __this_dir__ / "data/opencefadb-config.ttl"

    @property
    def rdf_store(self):
        """The database's main RDF store."""
        return self.main_rdf_store

    def add_hdf_infile_index(self):
        if hasattr(self.main_rdf_store, 'graph'):
            from .utils import build_infile_index_via_parents_for_graph
            idx_g = build_infile_index_via_parents_for_graph(self.main_rdf_store.graph, include_rootgroup=True)
            for s, p, o in tqdm.tqdm(idx_g, desc="Uploading HDF5 infile index triples", unit=" triples"):
                self.main_rdf_store.upload_triple((s, p, o))
        else:
            # is graphdb?!
            from .utils import build_infile_index_via_parents_for_graphdb
            build_infile_index_via_parents_for_graphdb(self.main_rdf_store)

    def download_cad_file(self, target_directory: Union[str, pathlib.Path], exist_ok=False):
        """Queries the RDF database for the iges cad file"""
        query_result = sparql_templates.fan.SELECT_FAN_CAD_FILE.execute(self.main_rdf_store)
        bindings = query_result.data
        n_bindings = len(bindings)
        if n_bindings != 1:
            raise ValueError(f"Expected one CAD file, got {n_bindings}")
        download_url = bindings["downloadURL"][0]
        _guess_filenames = download_url.rsplit("/", 1)[-1]
        target_directory = pathlib.Path(target_directory)
        target_filename = target_directory / _guess_filenames
        if target_filename.exists() and exist_ok:
            return target_filename
        return download_file(download_url, target_directory / _guess_filenames)

    # database specific methods:
    def get_fan_properties(self):
        from .sparql_templates.fan import SELECT_FAN_PROPERTIES
        res = SELECT_FAN_PROPERTIES.execute(self.main_rdf_store)
        return res.data  # {str(it["parameter"]): it for it in res.data.to_dict("records")}

    def get_fan_property(
            self,
            property_standard_name_uri: str
    ) -> NumericalVariable:
        """Gets a specific fan property from the fan curve data for the given rotational speed (in rpm) with the given tolerance."""
        q = sparql_templates.fan.get_fan_property(property_standard_name_uri)
        res = q.execute(self.main_rdf_store)
        return parse_to_entity(
            res.data,
            property_standard_name_uri,
            NumericalVariable
        )

    def get_operating_point_observations(
            self,
            operating_point_standard_names,
            standard_name_of_rotational_speed: str,
            n_rot_speed_rpm: float,
            n_rot_tolerance: float = 0.05
    ) -> List[Observation]:
        """Gets the fan curve data for the given rotational speed (in rpm) with the given tolerance
        and infer the OperatingPoint (SemanticOperatingPoint) objects from it

        ."""
        return get_operating_point_observations(
            self.main_rdf_store,
            operating_point_standard_names,
            standard_name_of_rotational_speed,
            n_rot_speed_rpm=n_rot_speed_rpm,
            n_rot_tolerance=n_rot_tolerance
        )


def get_operating_point_observations(
        rdf_store: h5cat.RDFStore,
        operating_point_standard_names,
        standard_name_of_rotational_speed,
        n_rot_speed_rpm: float,
        n_rot_tolerance: float = 0.05
):
    n_rot_range = tuple(
        [((1 - n_rot_tolerance) * n_rot_speed_rpm) / 60.,
         ((1 + n_rot_tolerance) * n_rot_speed_rpm) / 60.]
    )
    # TODO: this is not entirely correct, as the rotational speed dataset may have different units than rpm!
    q_n_rot = sparql_templates.hdf.find_dataset_for_standard_name(standard_name_of_rotational_speed, n_rot_range)
    standard_names_of_interest = operating_point_standard_names

    mean_data = defaultdict(dict)

    res_n_rot = q_n_rot.execute(rdf_store)

    if len(res_n_rot.data) > 0:
        for dataset_uri in tqdm.tqdm(res_n_rot.data["dataset"], desc="Processing operating point datasets",
                                     unit=" datasets"):
            q = sparql_templates.hdf.find_hdf5_file_for_dataset(dataset_uri)
            res_hdf_file = q.execute(rdf_store)
            hdf_file_uri = res_hdf_file.data.iloc[0]["hdfFile"]

            for target_standard_name in standard_names_of_interest:
                q = sparql_templates.hdf.find_dataset_in_file_by_standard_name(hdf_file_uri, target_standard_name)
                res = q.execute(rdf_store)
                if len(res) > 1:
                    raise ValueError(f"Expected one dataset for standard name {target_standard_name}, got {len(res)}")
                if len(res) == 0:
                    print(f"  no data for {target_standard_name}")
                    continue

                target_data = res.data.iloc[0]
                units_term = target_data.get("units")
                if units_term:
                    try:
                        units = units_term.toPython() if units_term is not None else None
                    except Exception:
                        units = str(units_term) if units_term is not None else None
                    unit_entity = get_unit_entity(units, rdf_store)
                else:
                    unit_entity = None
                standard_name_entity = get_standard_name_entity(target_standard_name, rdf_store)
                _data = {
                    "hasNumericalValue": target_data.get("value"),
                    "hasUnit": unit_entity,
                    "hasStandardName": standard_name_entity,
                    "hasSymbol": target_data.get("hasSymbol"),
                    "label": target_data.get("label"),
                    "altLabel": target_data.get("altLabel"),
                    "hasMinimumValue": target_data.get("hasMinimumValue"),
                    "hasMaximumValue": target_data.get("hasMaximumValue")
                }
                # find standard deviation counterpart:
                std_standard_name = target_standard_name.replace("arithmetic_mean_of", "standard_deviation_of")
                q = sparql_templates.hdf.find_dataset_in_file_by_standard_name(
                    dataset_uri,
                    std_standard_name,
                )
                res = q.execute(rdf_store)
                if len(res) == 1:
                    value = float(res.data["value"][0])
                    _data["hasUncertaintyDeclaration"] = StandardMU(has_standard_uncertainty=value)

                sn = NumericalVariable.model_validate(remove_none(_data))
                mean_data[dataset_uri][target_standard_name] = sn
                mean_data[dataset_uri]["hdf_uri"] = hdf_file_uri

    observations = []
    for k, v in mean_data.items():
        _hdf_uri = v.pop("hdf_uri")
        results = [ssn.Result(has_numerical_variable=nv) for nv in v.values()]
        hdf_dist = dcat.Distribution(
            id=_hdf_uri,
            media_type="application/x-hdf5",
            download_URL=_hdf_uri
        )
        observation = sosa.Observation(
            has_feature_of_interest=FAN_OPERATING_POINT,
            has_result=results,
            hadPrimarySource=hdf_dist
        )
        observations.append(observation)

    return observations


def _download_catalog(
        version: Optional[str] = None,
        target_directory: Optional[str] = None,
        sandbox: bool = False) -> Optional[pathlib.Path]:
    """download initial config"""
    if target_directory is None:
        target_directory = pathlib.Path.cwd()

    if sandbox:
        base_url = SANDBOX_BASE_URL
        dotenv.load_dotenv(pathlib.Path.cwd() / ".env", override=True)
        access_token = os.getenv("ZENODO_SANDBOX_API_TOKEN", None)
        config_filename = "opencefadb-config-sandbox.ttl"
    else:
        access_token = None
        base_url = PRODUCTION_BASE_URL
        config_filename = "opencefadb-catalog.ttl"
    pathlib.Path(target_directory).mkdir(parents=True, exist_ok=True)

    if version is not None and version.lower().strip() == "latest":
        version = None

    # if version is None, get the latest version of the zenodo record and download the file test.ttl, else get the specific version of the record:
    res = requests.get(base_url, params={'access_token': access_token} if sandbox else {})
    if res.status_code != 200:
        opencefa_print(f"Error: could not retrieve Zenodo record: {res.status_code}")
        raise ValueError(f"Error: could not retrieve Zenodo record: {res.status_code}")

    if version is None:
        opencefa_print("Searching for the latest version...")
        links = res.json().get("links", {})
        latest_version_url = links.get("latest", None)
        if latest_version_url is None:
            opencefa_print("Error: could not retrieve latest version URL from Zenodo record.")
            raise ValueError("Error: could not retrieve latest version URL from Zenodo record.")
        res = requests.get(latest_version_url, params={'access_token': access_token} if sandbox else {})
        if res.status_code != 200:
            opencefa_print(f"Error: could not retrieve latest version record from Zenodo: {res.status_code}")
            raise ValueError(f"Error: could not retrieve latest version record from Zenodo: {res.status_code}")
        detected_version = res.json()["metadata"]["version"]

        if sandbox:
            target_filename = pathlib.Path(
                target_directory) / f"opencefadb-catalog-sandbox-{detected_version.replace('.', '-')}.ttl"
        else:
            target_filename = pathlib.Path(
                target_directory) / f"opencefadb-catalog-{detected_version.replace('.', '-')}.ttl"

        for file in res.json().get("files", []):
            if file["key"] == config_filename:
                opencefa_print(f"downloading version {detected_version}...")
                file_res = requests.get(file["links"]["self"], params={'access_token': access_token} if sandbox else {})
                with open(target_filename, "wb") as f:
                    f.write(file_res.content)
                opencefa_print(f"Downloaded latest OpenCeFaDB config file to '{target_filename}'.")
                return target_filename
        opencefa_print(
            f"Error: Could not find config the file {config_filename} in the latest version of the Zenodo record.")
        raise ValueError(
            f"Error: Could not find config the file {config_filename} in the latest version of the Zenodo record.")

    opencefa_print(f"Searching for version {version}...")
    # a specific version is given:
    found_hit = None
    version_hits = requests.get(
        res.json()["links"]["versions"], params={'access_token': access_token} if sandbox else {}
    ).json()["hits"]["hits"]
    for hit in version_hits:
        if hit["metadata"]["version"] == version:
            found_hit = hit
            break
    if not found_hit:
        opencefa_print(f"Error: could not find version {version} in Zenodo record.")
        raise ValueError(f"Error: could not find version {version} in Zenodo record.")
    res_version = requests.get(found_hit["links"]["self"], params={'access_token': access_token} if sandbox else {})
    for file in res_version.json()["files"]:
        if file["key"] == config_filename:
            opencefa_print(f"Downloading version {version}...")
            file_res = requests.get(file["links"]["self"], params={'access_token': access_token} if sandbox else {})

            if sandbox:
                target_filename = pathlib.Path(
                    target_directory) / f"opencefadb-config-sandbox-{version.replace('.', '-')}.ttl"
            else:
                target_filename = pathlib.Path(
                    target_directory) / f"opencefadb-config-{version.replace('.', '-')}.ttl"

            with open(target_filename, "wb") as f:
                f.write(file_res.content)
            opencefa_print(f"Downloaded OpenCeFaDB config file version {version} to '{target_filename}'.")
            return target_filename
    opencefa_print(f"Error: could not find {config_filename} in the specified version.")
    return None



import pathlib
from typing import Union

import rdflib
import requests

from .url import _parse_url


def delete_repository(*,
                      repository_id: str,
                      host="http://localhost",
                      port=7200,
                      auth=(None, None)):
    """Delete a repository from a GraphDB instance.

    Equivalent to "curl -X DELETE localhost:7200/rest/repositories/<repository_id>"
    """
    response = requests.delete(f"{_parse_url(host)}:{port}/rest/repositories/{repository_id}", auth=auth)
    response.raise_for_status()
    return response


def create_repository(
        config_filename: Union[str, pathlib.Path],
        host="localhost",
        port=7200,
        exist_ok=True) -> str:
    """Create a repository in a GraphDB instance.

    Equivalent to `curl -X POST localhost:7201/rest/repositories -H "Content-Type: multipart/form-data" -F "config=@test-repo-config.ttl"`
    """
    config_filename = pathlib.Path(config_filename)
    assert config_filename.exists(), f"Configuration file '{config_filename}' does not exist"
    repoID = None
    g = rdflib.Graph()
    g.parse(config_filename, format="ttl")
    for res in g.query("""
    PREFIX rep: <http://www.openrdf.org/config/repository#>

    SELECT ?repoID
    WHERE {
        ?repo rep:repositoryID ?repoID .
    }
    """):
        repoID = str(res[0])

    if not config_filename.exists():
        raise FileNotFoundError(f"Configuration file '{config_filename}' does not exist")
    if not config_filename.suffix == '.ttl':
        raise ValueError("Configuration file must be a Turtle file (.ttl)")

    uri = f"{_parse_url(host)}:{port}"
    existing_repos = requests.get(f"{uri}/rest/repositories").json()
    if repoID in [e["id"] for e in existing_repos] and exist_ok:
        return repoID

    with open(config_filename, 'rb') as config_file:
        files = {'config': config_file}
        response = requests.post(f"{uri}/rest/repositories", files=files)
        response.raise_for_status()
    return repoID



from rdflib import Graph
from rdflib.plugins.stores.sparqlstore import SPARQLStore


def select_query(repo, sparql_query: str):
    store = SPARQLStore(auth=repo["auth"])
    store.open(repo["uri"])
    g = Graph(store=store, identifier=None, bind_namespaces="none")
    return g.query(sparql_query)


def ask_query():
    pass



import logging

import requests

logger = logging.getLogger("opencefadb")


def _get_request(*args, **kwargs):
    try:
        return requests.get(*args, **kwargs)
    except requests.exceptions.ConnectionError as e:
        logger.error(f"Connection error: {e}")
        raise e


def _post_request(*args, **kwargs):
    try:
        return requests.post(*args, **kwargs)
    except requests.exceptions.ConnectionError as e:
        logger.error(f"Connection error: {e}")
        raise e



from .templates import dcat
from .templates import hdf
from .templates import m4i
from .templates import person
from .templates import sensor
from .templates import qudt

MEASUREMENT_HDF_FILE_CONTENT_SHACL_DEFINITIONS = (
    # dcat.MINIMUM_DATASET_SHACL,
    hdf.NUMERIC_DATASETS_SHALL_HAVE_UNIT_AND_KIND_OF_QUANTITY,
    hdf.SHALL_HAVE_CREATOR,
    hdf.SHALL_HAVE_CREATED_DATE,
    hdf.SHALL_HAVE_VALID_ATTRIBUTION,
    hdf.HDF_FILE_SHALL_HAVE_STANDARD_NAME_TABLE,
    m4i.NUMERICAL_VARIABLE_HAS_LABEL,
    person.PERSON_SHACL,
    sensor.SHALL_HAVE_WELL_DESCRIBED_SSN_SENSOR,
    qudt.UNIT_NODE_SHALL_HAVE_KIND_OF_QUANTITY
)



NUMERIC_DATASETS_SHALL_HAVE_UNIT_AND_KIND_OF_QUANTITY = '''@prefix sh:   <http://www.w3.org/ns/shacl#> .
@prefix hdf:  <http://purl.allotrope.org/ontologies/hdf5/1.8#> .
@prefix m4i:  <http://w3id.org/nfdi4ing/metadata4ing#> .
@prefix ex:   <http://example.org/ns#> .

ex:NumericDatasetsMustHaveUnitAndKindOfQuantity
  a sh:NodeShape ;
  sh:targetClass hdf:Dataset ;
  sh:sparql [
    a sh:SPARQLConstraint ;
    sh:message "Numeric hdf:Dataset (H5T_INTEGER or H5T_FLOAT) must have at least one m4i:hasUnit and at least one m4i:hasKindOfQuantity; all values for both properties must be IRIs." ;
    sh:select """
      SELECT ?this WHERE {
        ?this a hdf:Dataset .

        # Only consider numeric datasets
        FILTER EXISTS {
          ?this hdf:datatype ?dt .
          FILTER (?dt IN (hdf:H5T_INTEGER, hdf:H5T_FLOAT))
        }

        # Violation if either requirement fails:
        #  - missing property entirely, OR
        #  - any value is not an IRI
        FILTER (
          # hasUnit missing OR has a non-IRI value
          !EXISTS { ?this m4i:hasUnit ?u . } ||
          EXISTS {
            ?this m4i:hasUnit ?u .
            FILTER ( !isIRI(?u) )
          } ||

          # hasKindOfQuantity missing OR has a non-IRI value
          !EXISTS { ?this m4i:hasKindOfQuantity ?kq . } ||
          EXISTS {
            ?this m4i:hasKindOfQuantity ?kq .
            FILTER ( !isIRI(?kq) )
          }
        )
      }
    """ ;
  ] .

'''

SHALL_HAVE_VALID_ATTRIBUTION = '''@prefix sh:   <http://www.w3.org/ns/shacl#> .
@prefix prov: <http://www.w3.org/ns/prov#> .
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix schema: <https://schema.org/> .
@prefix ex:   <http://example.org/ns#> .

ex:WasAttributedToMustReferenceAgent
    a sh:NodeShape ;
    sh:targetSubjectsOf prov:wasAttributedTo ;

    sh:property [
        sh:path prov:wasAttributedTo ;
        sh:message "prov:wasAttributedTo must point to a prov:Agent, prov:Person, foaf:Organization, prov:Organization, schema:Organization or schema:Person." ;

        # Allow any of the valid agent classes:
        sh:or (
            [ sh:class prov:Agent ]
            [ sh:class prov:Person ]
            [ sh:class foaf:Organization ]
            [ sh:class schema:Person ]
            [ sh:class schema:Organization ]
            [ sh:class prov:Organization ]
        ) ;
    ] .
'''

SHALL_HAVE_CREATED_DATE = '''@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix hdf: <http://purl.allotrope.org/ontologies/hdf5/1.8#> .
@prefix ex: <http://example.org/ns#> .

ex:HDFFileCreatedShape
    a sh:NodeShape ;
    sh:targetClass hdf:File ;                # apply only to hdf:File instances
    sh:property [
        sh:path dcterms:created ;            # must have this property
        sh:or (                             # accept either xsd:date or xsd:dateTime
            [ sh:datatype xsd:date ]
            [ sh:datatype xsd:dateTime ]
        ) ;
        sh:minCount 1 ;                      # at least one occurrence
        sh:maxCount 1 ;                      # optional but recommended
        sh:message "Each hdf:File must have exactly one dcterms:created value of type xsd:date." ;
    ] .
'''

SHALL_HAVE_CREATOR = '''@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix hdf: <http://purl.allotrope.org/ontologies/hdf5/1.8#> .
@prefix ex: <http://example.org/ns#> .

ex:HDFFileCreatorShape
    a sh:NodeShape ;
    sh:targetClass hdf:File ;
    sh:property [
        sh:path dcterms:creator ;
        sh:minCount 1 ;
        sh:message "Each hdf:File must have at least one dcterms:creator which is either an IRI or a prov:Person." ;
        sh:sparql [
            a sh:SPARQLConstraint ;
            sh:message "Each dcterms:creator must be either an IRI or a prov:Person." ;
            sh:select """
                SELECT ?this WHERE {
                  ?this dcterms:creator ?creator .

                  FILTER (
                    !isIRI(?creator) &&
                    NOT EXISTS { ?creator a prov:Person . }
                  )
                }
            """ ;
        ] ;
    ] .
    
'''

HDF_FILE_SHALL_HAVE_STANDARD_NAME_TABLE = '''@prefix sh:    <http://www.w3.org/ns/shacl#> .
@prefix ssno:  <https://matthiasprobst.github.io/ssno#> .
@prefix hdf:  <http://purl.allotrope.org/ontologies/hdf5/1.8#> .

ssno:HdfFileUsesStandardNameTableShape
    a sh:NodeShape ;
    sh:targetClass hdf:File ;
    sh:property [
        sh:path ssno:usesStandardNameTable ;
        sh:class ssno:StandardNameTable ;                 # object must be a StandardNameTable
        sh:minCount 1 ;
        sh:maxCount 1 ;                                # exactly one
        sh:message "A hdf:File file must use exactly one ssno:StandardNameTable via ssno:usesStandardNameTable."@en ;
    ] .
'''



import pathlib

import requests
from ontolutils.ex import prov
from ontolutils.ex.dcat import Dataset, Distribution
from ontolutils.ex.foaf import Agent
from h5rdmtoolbox.repository.zenodo import ZenodoRecord
ZENODO_API = "https://zenodo.org/api/records"

_EXT_MAP = {
    "csv": "text/csv",
    "tsv": "text/tab-separated-values",
    "json": "application/json",
    "jsonld": "application/ld+json",
    "ttl": "text/turtle",
    "hdf5": "application/x-hdf5",
    "h5": "application/x-hdf5",
    "nc": "application/x-netcdf",
    "zip": "application/zip",
    "iges": "model/iges",
    "igs": "model/iges",
    "md": "text/markdown",
    "txt": "text/plain",
    "xml": "application/xml",
    "rdf": "application/rdf+xml",
}


def _spdx_checksum_from_zenodo(checksum_str: str):
    """
    Convert Zenodo's 'md5:abcdef...' string to an SPDX checksum object.
    """
    if not checksum_str:
        return None
    if ":" in checksum_str:
        algo, value = checksum_str.split(":", 1)
    else:
        algo, value = "md5", checksum_str  # fallback

    algo_map = {
        "md5": "https://spdx.org/rdf/terms#checksumAlgorithm_md5",
        "sha1": "https://spdx.org/rdf/terms#checksumAlgorithm_sha1",
        "sha256": "https://spdx.org/rdf/terms#checksumAlgorithm_sha256",
        "sha512": "https://spdx.org/rdf/terms#checksumAlgorithm_sha512",
    }
    algo_iri = algo_map.get(algo.lower())
    return {
        "algorithm": algo_iri if algo_iri else algo,
        "checksumValue": value,
    }


def _zenodo_files(record):
    """
    Return a list of file dicts from a Zenodo record, handling common layouts.
    """
    files = record.get("files")
    if isinstance(files, list):  # modern layout
        return files
    # Older layouts sometimes keep files in metadata (rare for /records)
    md_files = record.get("metadata", {}).get("files", {})
    if isinstance(md_files, dict) and "files" in md_files:
        return md_files["files"]
    return []


def zenodo_record_to_dcat(record: dict) -> dict:
    """
    Translate a Zenodo record (as returned by /api/records) into a DCAT Dataset JSON-LD object.
    Includes dcat:Distribution entries with SPDX checksums for each file.
    """
    md = record.get("metadata", {})
    license_ = md.get("license").get("id") or None
    creators = md.get("creators") or []
    keywords = md.get("keywords") or []
    doi = md.get("doi")
    landing = record.get("links", {}).get("self")

    dcat_creators = []
    for c in creators:
        name = c.get("name") or c.get("affiliation")
        orcid = c.get("orcid") or ""
        person = {
            "id": f"https://orcid.org/{orcid.replace('https://orcid.org/', '')}",
            "name": name,
            "orcid": orcid,
            "affiliation": {
                "name": c.get("affiliation")
            }
        }
        dcat_creators.append(prov.Person.model_validate(person))

    def _parse_media_type(filename_suffix: str) -> str:
        ext = filename_suffix.rsplit('.', 1)[-1].lower()
        return _EXT_MAP.get(ext, "application/octet-stream")

    # Build distributions from files
    distributions = []
    for f in _zenodo_files(record):
        key = f.get("key") or f.get("filename") or "file"
        doi = doi or str(record.get("doi"))
        title = key
        media_type = _parse_media_type(pathlib.Path(f.get("key")).suffix)
        size = f.get("size")
        links = f.get("links", {})
        access_url = links.get("self") or record.get("links", {}).get("files")
        download_url = links.get("self")
        checksum_obj = _spdx_checksum_from_zenodo(f.get("checksum"))

        dist = {
            "id": links["self"],
            "title": title,
            "accessURL": access_url,
            "downloadURL": download_url,
            "mediaType": media_type,
            "byteSize": size,
            "checksum": checksum_obj
        }
        dcat_dist = Distribution.model_validate(dist)
        distributions.append(dcat_dist)
    publisher = Agent(
        id="https://zenodo.org/",
        name="Zenodo",
    )
    # Build dataset
    dataset = {
        "id": record["links"]["doi"],
        "title": md.get("title"),
        "description": md.get("description"),
        "identifier": doi or str(record.get("id")),
        "language": md.get("language"),
        "issued": md.get("publication_date"),
        "modified": record.get("updated") or record.get("revision"),
        "publisher": publisher,
        "creator": dcat_creators,
        "keyword": keywords,
        "landingPage": landing,
        "license": license_,
        "accessRights": md.get("access_right"),
        "version": md.get("version"),
        "distribution": distributions,
    }
    dcat_dataset = Dataset.model_validate(
        dataset
    )
    return dcat_dataset


def get_latest_opencefadb_datasets(max_pages=10):
    """
    Fetch all *latest* Zenodo records tagged 'opencefadb' that are of type 'Dataset'.
    """
    params = {
        "q": "keywords:opencefadb AND resource_type.type:dataset",
        "size": 100,
        "page": 1,
        "all_versions": "false",
        "sort": "mostrecent",
    }

    all_records = []
    for page in range(1, max_pages + 1):
        params["page"] = page
        resp = requests.get(ZENODO_API, params=params, timeout=30)
        resp.raise_for_status()
        data = resp.json()
        hits = data.get("hits", {}).get("hits", [])
        if not hits:
            break
        all_records.extend(hits)
    return all_records


if __name__ == "__main__":
    # 1) Fetch latest opencefadb datasets
    records = get_latest_opencefadb_datasets(max_pages=20)

    # 2) Convert each to DCAT
    dcat_datasets = [zenodo_record_to_dcat(r) for r in records]

    # 3) Print a tiny summary and checksums
    print(f"Converted {len(dcat_datasets)} datasets to DCAT.\n")
    for ds in dcat_datasets:
        print("Dataset:")
        print(ds.serialize("ttl"))

        # for dist in ds.distribution:
        #     dist.download(dest_filename=dist.title)



from typing import Union, Sequence, Callable, Optional

import matplotlib.pyplot as plt
import numpy as np
from ontolutils import QUDT_UNIT
from ontolutils.ex import qudt
from ontolutils.ex.sosa import ObservationCollection
from ontolutils.ex.ssn import Observation, Result
from ontolutils.namespacelib import QUDT_KIND
from rdflib.namespace import split_uri
from ssnolib.m4i import NumericalVariable

import opencefadb


def _parse_unit(unit: Union[str, qudt.Unit]) -> str:
    if isinstance(unit, str):
        unit = qudt.Unit.get(unit)
    if unit.symbol is None:
        unit = unit.expand()
    return unit


Selector = Union[str, Callable[[Observation], NumericalVariable]]
LabelResolver = Callable[[NumericalVariable], str]


def resolve_selector(sel: Selector) -> Callable[[Observation], NumericalVariable]:
    if callable(sel):
        return sel
    if isinstance(sel, str):
        return lambda obs: _standard_name_selector(obs, sel)
    raise TypeError(f"Unsupported selector type: {type(sel)}")


def _get_label(variable: NumericalVariable) -> Optional[str]:
    if variable.label is None:
        return None
    return str(variable.label)


def _get_alt_label(variable: NumericalVariable) -> Optional[str]:
    if variable.altLabel is None:
        return None
    return str(variable.altLabel)


def _get_symbol(variable: NumericalVariable) -> Optional[str]:
    if variable.hasSymbol is None:
        return None
    return str(variable.hasSymbol)


def _get_standard_name(variable: NumericalVariable) -> Optional[str]:
    if isinstance(variable.hasStandardName, str):
        if variable.hasStandardName.startswith("http"):
            return split_uri(variable.hasStandardName)[-1]
        else:
            return str(variable.hasStandardName)
    elif variable.hasStandardName.standardName is not None:
        return str(variable.hasStandardName.standardName)
    return None


LABEL_SELECTION_MAPPER = {
    "label": _get_label,
    "symbol": _get_symbol,
    "altLabel": _get_alt_label,
    "standard_name": _get_standard_name,
}


class DefaultLabelResolver:
    LABEL_SELECTION_ORDER = [
        "label",
        "symbol",
        "altLabel",
        "standard_name",
    ]

    def __call__(self, variable: NumericalVariable) -> str:
        unit = None
        name = None
        for sel in self.LABEL_SELECTION_ORDER:
            lso = LABEL_SELECTION_MAPPER.get(sel, None)
            if lso is None:
                raise ValueError(f"Unknown label selection option: {sel}")
            name = lso(variable)
            if name is not None:
                break

        if variable.hasUnit is not None:
            if isinstance(variable.hasUnit, str):
                try:
                    qunit = _parse_unit(variable.hasUnit)
                    unit = qunit.symbol
                except Exception:
                    unit = str(variable.hasUnit)
            elif isinstance(variable.hasUnit, qudt.Unit):
                qunit = variable.hasUnit.expand()
                unit = qunit.symbol
            # elif variable.hasUnit.unit is not None:
            #     unit = str(variable.hasUnit.unit)

        if name is None:
            name = "?"
        if unit is None:
            unit = "?"
        return f"{name} / {unit}"


def _standard_name_selector(obs: Observation, standard_name: str) -> Optional[NumericalVariable]:
    wanted = str(standard_name).strip()
    is_iri = wanted.startswith("http://") or wanted.startswith("https://")

    if is_iri:
        for result in obs.hasResult:
            nv = result.hasNumericalVariable
            if nv:
                sn = nv.hasStandardName
                if isinstance(sn, str):
                    if str(sn) == wanted:
                        return nv
                elif sn and str(sn.id) == wanted:
                    return nv
    else:
        for result in obs.hasResult:
            nv = result.hasNumericalVariable
            if nv:
                sn = nv.hasStandardName
                if isinstance(sn, str):
                    continue
                if sn and sn.standardName == wanted:
                    return nv
    return None


class SemanticOperationPoint:
    EXPECTED_PRESSURE_QUANTITY_KINDS = [
        QUDT_KIND.Pressure,
        QUDT_KIND.StaticPressure,
        QUDT_KIND.TotalPressure,
    ]
    ROTATIONAL_FREQUENCY_QUANTITY_KIND = QUDT_KIND.RotationalVelocity
    VOLUME_FLOW_RATE_QUANTITY_KIND = QUDT_KIND.VolumeFlowRate

    def __init__(self, observation: Observation):
        self.observation = observation

    def scale(
            self,
            reference_rotational_frequency: NumericalVariable,
            keep_other_results:bool=True,
    ) -> "SemanticOperationPoint":
        """Scale the operating point by the provided numerical variable according to the affinity laws.

        Parameters
        ----------
        reference_rotational_frequency : NumericalVariable
            The reference rotational frequency to scale to.
        keep_other_results : bool, optional
            Whether to keep other results that are not scaled, by default True.
        """
        scaled_result_ids = []

        if not isinstance(reference_rotational_frequency, NumericalVariable):
            raise TypeError("reference_rotational_frequency must be a NumericalVariable")
        if not reference_rotational_frequency.is_kind_of_quantity(self.ROTATIONAL_FREQUENCY_QUANTITY_KIND):
            raise ValueError("reference_rotational_frequency must be of kind of quantity "
                             f"{self.ROTATIONAL_FREQUENCY_QUANTITY_KIND}: {reference_rotational_frequency}")

        observation_rotational_frequency = self.observation.get_numerical_variable_by_kind_of_quantity(
            self.ROTATIONAL_FREQUENCY_QUANTITY_KIND
        )
        if len(observation_rotational_frequency) != 1:
            print(observation_rotational_frequency)
            raise ValueError(
                f"Fan curve must have exactly one numerical variable of kind {self.ROTATIONAL_FREQUENCY_QUANTITY_KIND}")
        observation_rotational_frequency = observation_rotational_frequency[0].to_pint()
        for res in self.observation.hasResult:
            nv = res.hasNumericalVariable
            if nv is not None and nv.is_kind_of_quantity(self.ROTATIONAL_FREQUENCY_QUANTITY_KIND):
                scaled_result_ids.append(res.id)
                break

        if "turn" in observation_rotational_frequency._units:
            observation_rotational_frequency_pint = observation_rotational_frequency.to("turns/s")
        else:
            observation_rotational_frequency_pint = observation_rotational_frequency.to("1/s")

        vfr_numerical_variables = self.observation.get_numerical_variable_by_kind_of_quantity(
            self.VOLUME_FLOW_RATE_QUANTITY_KIND
        )
        if len(vfr_numerical_variables) != 1:
            raise ValueError(
                f"Fan curve must have exactly one numerical variable of kind {self.VOLUME_FLOW_RATE_QUANTITY_KIND}")
        for res in self.observation.hasResult:
            nv = res.hasNumericalVariable
            if nv is not None and nv.is_kind_of_quantity(self.ROTATIONAL_FREQUENCY_QUANTITY_KIND):
                scaled_result_ids.append(res.id)
                break

        pressure_numerical_variables = []
        for PKC in self.EXPECTED_PRESSURE_QUANTITY_KINDS:
            _pressure_variable = self.observation.get_numerical_variable_by_kind_of_quantity(
                PKC
            )
            if len(_pressure_variable) > 0:
                pressure_numerical_variables.extend(_pressure_variable)
            elif len(_pressure_variable) == 0:
                # consider units: # TODO this is a workaround for missing kind_of_quantity
                for result in self.observation.hasResult:
                    nv = result.hasNumericalVariable
                    if nv:
                        if isinstance(nv.hasUnit, str):
                            if str(nv.hasUnit) == str(QUDT_UNIT.PA):
                                if nv not in pressure_numerical_variables:
                                    pressure_numerical_variables.append(nv)
                                    scaled_result_ids.append(result.id)
                        elif str(nv.hasUnit.id) == str(QUDT_UNIT.PA):
                            if nv not in pressure_numerical_variables:
                                pressure_numerical_variables.append(nv)
                                scaled_result_ids.append(result.id)

        reference_rotational_frequency_pint = reference_rotational_frequency.to_pint()
        if "turn" in reference_rotational_frequency_pint._units:
            reference_rotational_frequency_pint = reference_rotational_frequency_pint.to("turns/s")
        else:
            reference_rotational_frequency_pint = reference_rotational_frequency_pint.to("1/s")

        scaled_variables = []
        for p in pressure_numerical_variables:
            scaled_pressure_value = p.hasNumericalValue * (
                    reference_rotational_frequency_pint / observation_rotational_frequency_pint) ** 2
            scaled_p = p.model_copy()
            scaled_p.hasNumericalValue = scaled_pressure_value.magnitude
            scaled_variables.append(scaled_p)

        for vfr in vfr_numerical_variables:
            scaled_vfr_value = vfr.hasNumericalValue * (
                    reference_rotational_frequency_pint / observation_rotational_frequency_pint)
            scaled_vfr = vfr.model_copy()
            scaled_vfr.hasNumericalValue = scaled_vfr_value.magnitude
            scaled_variables.append(scaled_vfr)

        scaled_variables.append(
            reference_rotational_frequency.model_copy()
        )

        if keep_other_results:
            for result in self.observation.hasResult:
                if result.id in scaled_result_ids:
                    continue
                nv = result.hasNumericalVariable
                if nv is not None:
                    scaled_variables.append(nv)

        scaled_observation = Observation(
            has_result=[
                Result(
                    has_numerical_variable=n
                )
                for n in scaled_variables
            ]
        )
        return SemanticOperationPoint(scaled_observation)


def get_result_by_standard_name(observation: Observation, name: str=None, iri:str=None):
        """Get all results matching the provided standard name or IRI."""
        if name is None and iri is None:
            raise ValueError("Either name or iri must be provided")
        for result in observation.hasResult:
            nv = result.hasNumericalVariable
            if nv is not None:
                sn = nv.hasStandardName
                if name is not None:
                    if isinstance(sn, str):
                        continue
                    if sn and sn.standardName == name:
                        return result
                elif iri is not None:
                    if isinstance(sn, str):
                        if str(sn) == iri:
                            return result
                    elif sn and str(sn.id) == iri:
                        return result
        return None


class SemanticFanCurve:
    """
    A view on an ObservationCollection that enables plotting based on two
    selected result variables. The collection is the single source of truth.
    """

    def __init__(self, collection: ObservationCollection):
        if collection is None:
            raise ValueError("collection must not be None")
        if not isinstance(collection, ObservationCollection):
            raise TypeError("collection must be an ObservationCollection")
        self.collection: ObservationCollection = collection

    def __len__(self):
        return len(self.collection.hasMember)

    def __repr__(self):
        return f"<SemanticFanCurve with {len(self)} observations>"

    @classmethod
    def from_observations(cls, observations: Sequence[Observation], **kwargs) -> "SemanticFanCurve":
        observations = list(observations)
        feature_of_interests = [
            obs.hasFeatureOfInterest for obs in observations if
            obs.hasFeatureOfInterest is not None
        ]
        # check if they all have the same foi
        if len(set(feature_of_interests)) > 1:
            raise ValueError("All observations must have the same feature of interest")
        foi = feature_of_interests[0] if len(feature_of_interests) > 0 else None
        for k, v in kwargs.copy().items():
            if v is None:
                kwargs.pop(k)
        oc = ObservationCollection(has_member=list(observations), **kwargs)
        if foi:
            oc.hasFeatureOfInterest = foi
        return cls(oc)

    def get_result_by_standard_name(self, name: str=None, iri:str=None):
        """Get all results matching the provided standard name or IRI."""
        if name is None and iri is None:
            raise ValueError("Either name or iri must be provided")
        results = []
        for obs in self.collection.hasMember:
            _res = get_result_by_standard_name(obs, name=name, iri=iri)
            if _res is not None:
                results.append(_res)
        return results

    def get_result_by_kind_of_quantity(self, koq: qudt.QuantityKind=None, iri: str=None):
        """Get all results matching the provided kind of quantity."""
        if koq is None and iri is None:
            raise ValueError("Either koq or iri must be provided")

        results = []
        for obs in self.collection.hasMember:
            for result in obs.hasResult:
                nv = result.hasNumericalVariable
                if nv is not None:
                    if koq is not None:
                        if nv.is_kind_of_quantity(koq):
                            results.append(result)
                    elif iri is not None:
                        if nv.hasKindOfQuantity is not None:
                            if isinstance(nv.hasKindOfQuantity, str):
                                if str(nv.hasKindOfQuantity) == iri:
                                    results.append(result)
                            elif str(nv.hasKindOfQuantity.id) == iri:
                                results.append(result)
        return results

    def get_xy(self, x: Selector, y: Selector):
        fx = resolve_selector(x)
        fy = resolve_selector(y)

        xs, ys = [], []
        for obs in self.collection.hasMember:
            try:
                xs.append(fx(obs))
                ys.append(fy(obs))
            except KeyError:
                # in der Toolbox: entweder skip + log, oder strict fail
                continue
        return xs, ys

    def serialize(self, format: str = "turtle", **kwargs) -> str:
        return self.collection.serialize(format=format, **kwargs)

    def scale(
            self,
            n: NumericalVariable,
            keep_other_results: bool = True,
    ):
        """Scale the fan curve by the provided numerical variable according to the affinity laws."""
        if len(self) == 0:
            raise ValueError("Cannot scale an empty fan curve.")
        scaled_members = [
            SemanticOperationPoint(observation).scale(n, keep_other_results=keep_other_results).observation for observation in self.collection.hasMember
        ]
        return SemanticFanCurve.from_observations(
            scaled_members,
            id=self.collection.id,
            hasFeatureOfInterest=self.collection.hasFeatureOfInterest,
        )
        # for observation in self.collection.hasMember:
        #     scaled_observations =
        #     for result in obs.hasResult:
        #         nv = result.hasNumericalVariable
        #         if nv is not None:
        #             if nv.has_numerical_value is not None:
        #                 scaled_value = nv.has_numerical_value * n.has_numerical_value
        #             else:
        #                 scaled_value = None
        #             scaled_nv = NumericalVariable(
        #                 id=nv.id,
        #                 label=nv.label,
        #                 hasStandardName=nv.hasStandardName,
        #                 has_numerical_value=scaled_value,
        #                 hasUnit=nv.hasUnit,
        #             )
        #             scaled_result = type(result)(
        #                 id=result.id,
        #                 hasNumericalVariable=scaled_nv,
        #             )
        #             scaled_results.append(scaled_result)
        #         else:
        #             scaled_results.append(result)
        #     scaled_obs = type(obs)(
        #         id=obs.id,
        #         hasResult=scaled_results,
        #         hasFeatureOfInterest=obs.hasFeatureOfInterest,
        #         hadPrimarySource=obs.hadPrimarySource,
        #     )
        #     scaled_members.append(scaled_obs)

    def _get_plotting_data(
            self,
            x: Selector,
            y: Selector,
            xlabel: Union[str, LabelResolver] = None,
            ylabel: Union[str, LabelResolver] = None,
            xsort: bool = True,
            ret_err: bool = False,
            raise_on_no_data_points=True
    ):
        resolved_x, resolved_y = self.get_xy(x, y)

        # strip None values in xs and ys (pair-wise)
        xy_filtered = [(xv, yv) for xv, yv in zip(resolved_x, resolved_y) if xv is not None and yv is not None]
        xs, ys = [t[0].hasNumericalValue for t in xy_filtered], [t[1].hasNumericalValue for t in xy_filtered]
        if len(xy_filtered) == 0:
            if raise_on_no_data_points:
                raise ValueError("No data points could be extracted.")
            else:
                return [], [], None, None, "x / ?", "y / ?"
        if xsort:
            xy = sorted(zip(xs, ys), key=lambda t: t[0])
            xs, ys = [t[0] for t in xy], [t[1] for t in xy]

        if ret_err:
            # hasUncertaintyDeclaration may be null!
            xerr = []
            for t in xy_filtered:
                if t[0].hasUncertaintyDeclaration is not None:
                    xerr.append(t[0].hasUncertaintyDeclaration.has_standard_uncertainty)
                else:
                    xerr.append(np.nan)
            yerr = []
            for t in xy_filtered:
                if t[1].hasUncertaintyDeclaration is not None:
                    yerr.append(t[1].hasUncertaintyDeclaration.has_standard_uncertainty)
                else:
                    yerr.append(np.nan)
        else:
            xerr, yerr = None, None

        _xlabel = "x / ?"
        if xlabel is None:
            _xlabels = [DefaultLabelResolver()(it) for it in resolved_x if it is not None]
            for it in _xlabels:
                if it is not None:
                    _xlabel = it
                    break
        else:
            if isinstance(xlabel, str):
                _xlabel = xlabel
            else:
                _xlabel = "x / ?"

        _ylabel = "y / ?"
        if ylabel is None:
            _ylabels = [DefaultLabelResolver()(it) for it in resolved_y if it is not None]
            for it in _ylabels:
                if it is not None:
                    _ylabel = it
                    break
        else:
            if isinstance(ylabel, str):
                _ylabel = ylabel
            else:
                _ylabel = "y / ?"

        return xs, ys, xerr, yerr, _xlabel, _ylabel

    def plot(
            self,
            x: Selector,
            y: Selector,
            xlabel: Union[str, LabelResolver] = None,
            ylabel: Union[str, LabelResolver] = None,
            xsort: bool = True,
            raise_on_no_data_points: bool = True,
            **kwargs
    ):
        verbose = kwargs.pop("verbose", None)
        ax = kwargs.pop("ax", None)
        if ax is None:
            ax = plt.gca()

        xs, ys, xerr, yerr, _xlabel, _ylabel = self._get_plotting_data(x, y, xlabel, ylabel, xsort,
                                                                       raise_on_no_data_points=raise_on_no_data_points)

        if ys:
            ax.plot(xs, ys, **kwargs)
            plt.xlabel(_xlabel)
            plt.ylabel(_ylabel)
        return ax

    def errorbar(
            self,
            x: Selector,
            y: Selector,
            xlabel: Union[str, LabelResolver] = None,
            ylabel: Union[str, LabelResolver] = None,
            xsort: bool = True,
            raise_on_no_data_points: bool = True,
            **kwargs
    ):
        verbose = kwargs.pop("verbose", None)
        ax = kwargs.pop("ax", None)
        if ax is None:
            ax = plt.gca()
        opencefadb.opencefa_print("Obtaining plotting data...", verbose=verbose)
        xs, ys, xs_err, ys_err, _xlabel, _ylabel = self._get_plotting_data(x, y, xlabel, ylabel, xsort, ret_err=True,
                                                                           raise_on_no_data_points=raise_on_no_data_points)
        opencefadb.opencefa_print("... done.", verbose=verbose)
        if ys:
            ax.errorbar(xs, ys, xerr=xs_err, yerr=ys_err, **kwargs)
            plt.xlabel(_xlabel)
            plt.ylabel(_ylabel)
        return ax



# import hashlib
# import logging
# import pathlib
# from abc import ABC, abstractmethod
# from typing import List
#
# import h5py
# from h5rdmtoolbox.catalog.stores import DataStore
# from h5rdmtoolbox.catalog.query.query import AbstractQuery
# from h5rdmtoolbox.database.hdfdb import FilesDB
#
# logger = logging.getLogger("opencefadb")
#
#
# class HDF5FileQuery(AbstractQuery, ABC):
#
#     def __init__(self, filter, objfilter=None, recursive=True, **kwargs):
#         self._filter = filter
#         self._objfilter = objfilter
#         self._recursive = recursive
#         self._kwargs = kwargs
#
#     @abstractmethod
#     def execute(self, filenames: List[pathlib.Path], *args, **kwargs):
#         pass
#
#
# class FindOneLazyHDFObject(HDF5FileQuery):
#     def execute(self, filenames: List[pathlib.Path], *args, **kwargs):
#         return FilesDB(filenames).find_one(flt=self._filter,
#                                            objfilter=self._objfilter,
#                                            recursive=self._recursive,
#                                            **kwargs)
#
#
# class FindManyLazyHDFObject(HDF5FileQuery):
#     def execute(self, filenames: List[pathlib.Path], *args, **kwargs):
#         return FilesDB(filenames).find(flt=self._filter,
#                                        objfilter=self._objfilter,
#                                        recursive=self._recursive, **kwargs)
#
#
# class HDF5FileDB(DataStore):
#
#     def __init__(self):
#         self._filenames = {}
#         self._expected_file_extensions = {".hdf", ".hdf5", ".h5"}
#
#     def __repr__(self):
#         return f"<{self.__class__.__name__} (n_files={len(self.filenames)})>"
#
#     @property
#     def filenames(self):
#         return self._filenames
#
#     def upload_file(self, filename) -> str:
#         filename = pathlib.Path(filename)
#         import hashlib
#         hash_obj = hashlib.sha256()
#         hash_obj.update(str(filename.resolve().absolute()).encode())
#
#         assert filename.exists(), f"File {filename} does not exist."
#         assert filename.suffix in self._expected_file_extensions, f"File type {filename.suffix} not supported"
#
#         h5_hash = h5hash(str(filename))
#
#         if h5_hash in self._filenames:
#             return h5_hash
#
#         logger.debug(f"Uploading file {filename} to {self}")
#         self._filenames[h5_hash] = filename.resolve().absolute()
#         return h5_hash
#
#     def execute_query(self, query: HDF5FileQuery):
#         return query.execute(self.filenames)
#
#
# def h5hash(filename: str) -> str:
#     """Generate a deterministic ID based on metadata and content."""
#     hash_obj = hashlib.sha256()
#
#     with h5py.File(filename, 'r') as f:
#         # Include metadata in the hash
#         for k, v in f.attrs.items():
#             hash_obj.update(k.encode())
#             hash_obj.update(v.encode())
#
#         # Include content in the hash
#         def hash_group(group):
#             """Recursively hash the datasets in a group."""
#             for key in group:
#                 item = group[key]
#                 if isinstance(item, h5py.Group):
#                     hash_group(item)
#                 elif isinstance(item, h5py.Dataset):
#                     hash_obj.update(item[()].tobytes())  # Update hash with dataset content
#
#         hash_group(f)
#
#     return hash_obj.hexdigest()



from typing import Union

import rdflib
from h5rdmtoolbox import catalog as h5cat

SELECT_FAN_PROPERTIES = h5cat.SparqlQuery(
    query="""PREFIX m4i: <http://w3id.org/nfdi4ing/metadata4ing#>

SELECT ?parameter ?property ?value
WHERE {
  <https://www.wikidata.org/wiki/Q131549102> m4i:hasParameter ?parameter .
  ?parameter ?property ?value .
}

ORDER BY ?parameter ?property
""",
    description="Selects all properties of the fan")


def get_fan_property(standard_name_uri: Union[str, rdflib.URIRef]):
    """Returns a SPARQL query that selects all properties of the fan parameter with the given standard name URI."""
    if isinstance(standard_name_uri, rdflib.URIRef):
        standard_name_uri = str(standard_name_uri)
    query_str = f"""PREFIX m4i: <http://w3id.org/nfdi4ing/metadata4ing#>
PREFIX ssno: <https://matthiasprobst.github.io/ssno#>

SELECT ?property ?value
WHERE {{
  <https://www.wikidata.org/wiki/Q131549102> m4i:hasParameter ?parameter .
  ?parameter ssno:hasStandardName <{standard_name_uri}> .
  ?parameter ?property ?value .
}}
ORDER BY ?property
"""
    return h5cat.SparqlQuery(
        query=query_str,
        description=f"Selects all properties of the fan parameter with standard name {standard_name_uri}"
    )


SELECT_FAN_CAD_FILE = h5cat.SparqlQuery(
    query="""
PREFIX schema: <http://schema.org/>
PREFIX dcterms: <http://purl.org/dc/terms/>
PREFIX dcat: <http://www.w3.org/ns/dcat#>

SELECT DISTINCT ?downloadURL
WHERE {
  <https://www.wikidata.org/wiki/Q131549102> dcterms:hasPart ?part .
  ?part (dcat:distribution|schema:distribution|dcterms:hasPart) ?distribution .
  ?distribution (dcat:downloadURL|schema:downloadUrl|dcterms:identifier) ?downloadURL .
  FILTER(BOUND(?downloadURL))
}

LIMIT 10
""",
    description="Selects the CAD file for the fan")


# def construct_data_based_on_standard_name_based_search(
#         target_standard_name_uris: list[str]
# ) -> h5cat.SparqlQuery:
#     """
#     Finds data for multiple target standard name URIs per HDF file where the file contains a
#     dataset with standard name `conditional_standard_name_uri` whose value is within the given range.
#
#     :param target_standard_name_uris: list of standard name URIs to query for
#     :return:
#     """
#     values_block = " ".join(f"<{uri}>" for uri in target_standard_name_uris)
#     query_str = f"""
# PREFIX hdf: <http://purl.allotrope.org/ontologies/hdf5/1.8#>
# PREFIX ssno: <https://matthiasprobst.github.io/ssno#>
# PREFIX m4i: <http://w3id.org/nfdi4ing/metadata4ing#>
# PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
# PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
#
# SELECT ?hdfFile ?dataset ?value ?units ?standardName ?datasetName ?hasSymbol ?altLabel ?label
# WHERE {{
#   VALUES ?standardName {{ {values_block} }}
#   ?dataset a hdf:Dataset ;
#            ssno:hasStandardName ?standardName .
#
#   OPTIONAL {{ ?dataset hdf:value ?value }}
#   OPTIONAL {{ ?dataset hdf:name ?datasetName }}
#   OPTIONAL {{ ?dataset m4i:hasSymbol ?hasSymbol }}
#
#   OPTIONAL {{ ?dataset ssno:unit ?unit1 }}
#   OPTIONAL {{ ?dataset m4i:hasUnit ?unit2 }}
#   OPTIONAL {{ ?dataset rdfs:label ?label }}
#   OPTIONAL {{ ?dataset skos:altLabel ?altLabel }}
#   BIND(COALESCE(?unit1, ?unit2) AS ?units)
#
#   ?hdfFile (hdf:rootGroup/hdf:member*)* ?dataset .
#   ?hdfFile a hdf:File .
# }}
# ORDER BY ?hdfFile ?dataset
# """
#     description = f"Selects datasets with standard names {target_standard_name_uris}"
#     return h5cat.SparqlQuery(
#         query=query_str,
#         description=description
#     )


# def construct_data_based_on_standard_name_based_search_and_range_condition(
#         target_standard_name_uris: list[str],
#         conditional_standard_name_uri: str,
#         condition_range: tuple[float, float]
# ) -> h5cat.SparqlQuery:
#     values_block = " ".join(f"<{uri}>" for uri in target_standard_name_uris)
#
#     query_str = f"""
# PREFIX hdf: <http://purl.allotrope.org/ontologies/hdf5/1.8#>
# PREFIX ssno: <https://matthiasprobst.github.io/ssno#>
# PREFIX m4i: <http://w3id.org/nfdi4ing/metadata4ing#>
# PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
# PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
# PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
#
# SELECT ?hdfFile ?dataset ?value ?units ?standardName ?datasetName ?hasSymbol ?altLabel ?label
# WHERE {{
#   VALUES ?standardName {{ {values_block} }}
#
#   ?hdfFile a hdf:File ;
#            hdf:rootGroup ?root .
#
#   # Traverse file structure via groups (matches your model)
#   ?root hdf:member* ?group .
#   ?group a hdf:Group ;
#          hdf:member ?dataset .
#
#   ?dataset a hdf:Dataset ;
#            ssno:hasStandardName ?standardName .
#
#   ?dataset hdf:value ?value .
#   ?dataset hdf:name ?datasetName .
#   OPTIONAL {{ ?dataset m4i:hasSymbol ?hasSymbol }}
#   OPTIONAL {{ ?dataset ssno:unit ?unit1 }}
#   OPTIONAL {{ ?dataset m4i:hasUnit ?unit2 }}
#   OPTIONAL {{ ?dataset rdfs:label ?label }}
#   OPTIONAL {{ ?dataset skos:altLabel ?altLabel }}
#   BIND(COALESCE(?unit1, ?unit2) AS ?units)
#
#   # Condition: file must contain a dataset with the conditional standard name in the numeric range
#   FILTER EXISTS {{
#     ?root hdf:member* ?g2 .
#     ?g2 hdf:member ?rotDataset .
#     ?rotDataset a hdf:Dataset ;
#                 ssno:hasStandardName <{conditional_standard_name_uri}> ;
#                 hdf:value ?conditionValue .
#     FILTER(
#       xsd:double(?conditionValue) >= {condition_range[0]}
#       && xsd:double(?conditionValue) <= {condition_range[1]}
#     )
#   }}
# }}
# ORDER BY ?hdfFile ?dataset
# """
#     description = (f"Selects datasets with standard names {target_standard_name_uris} within range "
#                    f"{condition_range} of datasets with standard name {conditional_standard_name_uri}")
#     return h5cat.SparqlQuery(query=query_str, description=description)

#


# def find_dataset_value_in_same_group_by_other_standard_name(
#         dataset_a: str,
#         standard_name_b: str,
# ) -> h5cat.SparqlQuery:
#     query_str = f"""
# PREFIX hdf:  <http://purl.allotrope.org/ontologies/hdf5/1.8#>
# PREFIX ssno: <https://matthiasprobst.github.io/ssno#>
# PREFIX m4i:  <http://w3id.org/nfdi4ing/metadata4ing#>
# PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
# PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
#
# SELECT DISTINCT
#   ?hdfFile ?group ?dsA ?dsB ?value ?units ?label ?altLabel ?hasSymbol ?hasMinimumValue ?hasMaximumValue
# WHERE {{
#   BIND(<{dataset_a}> AS ?dsA)
#   BIND(<{standard_name_b}> AS ?stdB)
#
#   # 1) Find the group(s) that directly contain dataset A (no property paths)
#   ?group a hdf:Group ;
#          hdf:member ?dsA ;
#          hdf:member ?dsB .
#
#   # 2) Constrain dataset B early
#   ?dsB a hdf:Dataset ;
#        ssno:hasStandardName ?stdB ;
#        hdf:value ?value .
#
#   FILTER(?dsB != ?dsA)
#
#   # 3) Only now, resolve which file this group belongs to
#   ?hdfFile a hdf:File ;
#            hdf:rootGroup ?root .
#
#   {{
#     # group directly under root
#     ?root hdf:member ?group .
#   }}
#   UNION
#   {{
#     # group under subgroups
#     ?root hdf:member+ ?g1 .
#     ?g1  hdf:member ?group .
#   }}
#
#   OPTIONAL {{ ?dsB m4i:hasMinimumValue ?hasMinimumValue }}
#   OPTIONAL {{ ?dsB m4i:hasMaximumValue ?hasMaximumValue }}
#   OPTIONAL {{ ?dsB m4i:hasSymbol ?hasSymbol }}
#   OPTIONAL {{ ?dsB m4i:hasUnit ?units }}
#   OPTIONAL {{ ?dsB rdfs:label ?label }}
#   OPTIONAL {{ ?dsB skos:altLabel ?altLabel }}
# }}
# """
#     return h5cat.SparqlQuery(
#         query=query_str,
#         description=f"Find dataset with standard name {standard_name_b} in same group as {dataset_a}"
#     )


# def find_dataset_value_in_same_group_by_other_standard_name2(
#         dataset_a: str,
#         standard_name_b: str,
# ) -> h5cat.SparqlQuery:
#     query_str = f"""
# PREFIX hdf: <http://purl.allotrope.org/ontologies/hdf5/1.8#>
# PREFIX ssno: <https://matthiasprobst.github.io/ssno#>
# PREFIX m4i: <http://w3id.org/nfdi4ing/metadata4ing#>
# PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
#
# SELECT ?hdfFile ?group ?dsA ?dsB ?value ?units ?label ?altLabel ?hasSymbol ?hasMinimumValue ?hasMaximumValue
# WHERE {{
#   VALUES ?dsA  {{ <{dataset_a}> }}
#   VALUES ?stdB {{ <{standard_name_b}> }}
#
#   # file + group hierarchy
#   ?hdfFile a hdf:File ;
#            hdf:rootGroup ?root .
#
#   # find the group (under this file) that directly contains dataset A
#   ?root hdf:member* ?group .
#   ?group a hdf:Group ;
#          hdf:member ?dsA .
#
#   # find dataset B in the same group with the requested standard name
#   ?group hdf:member ?dsB .
#   ?dsB a hdf:Dataset ;
#        ssno:hasStandardName ?stdB .
#   ?dsB hdf:value ?value
#
#   OPTIONAL {{ ?dsB m4i:hasMinimumValue ?hasMinimumValue }}
#   OPTIONAL {{ ?dsB m4i:hasMaximumValue ?hasMaximumValue }}
#   OPTIONAL {{ ?dsB m4i:hasSymbol ?hasSymbol }}
#   OPTIONAL {{ ?dsB m4i:hasUnit ?units }}
#   OPTIONAL {{ ?dsB rdfs:label ?label }}
#   OPTIONAL {{ ?dsB skos:altLabel ?altLabel }}
#
#   FILTER(?dsB != ?dsA)
# }}
# """
#     return h5cat.SparqlQuery(
#         query=query_str,
#         description=f"Find dataset with standard name {standard_name_b} in same group as {dataset_a}"
#     )
#
#




import unittest

import h5rdmtoolbox as h5tbx
import pytest
import rdflib
from ontolutils.ex.qudt import Unit
from ontolutils.ex.sosa import ObservableProperty, Sensor
from ontolutils.ex.ssn import Accuracy, SystemCapability, MeasurementRange
from rdflib import DCTERMS, PROV

from opencefadb.validation import validate_hdf
from opencefadb.validation.shacl.templates.hdf import SHALL_HAVE_CREATED_DATE, \
    NUMERIC_DATASETS_SHALL_HAVE_UNIT_AND_KIND_OF_QUANTITY, \
    SHALL_HAVE_CREATOR
from opencefadb.validation.shacl.templates.sensor import SHALL_HAVE_WELL_DESCRIBED_SSN_SENSOR


class TestPlotting(unittest.TestCase):

    def test_hdf_numeric_datasets_shall_have_unit(self):
        with h5tbx.File() as h5:
            h5.create_dataset(
                "numeric_dataset_no_unit",
                data=[1, 2, 3, 4, 5],
                dtype='i4'
            )

        res = validate_hdf(
            hdf_source=h5.hdf_filename,
            shacl_data=NUMERIC_DATASETS_SHALL_HAVE_UNIT_AND_KIND_OF_QUANTITY
        )
        self.assertFalse(res.conforms)

        with h5tbx.File() as h5:
            ds = h5.create_dataset(
                "numeric_dataset_no_unit",
                data=[1, 2, 3, 4, 5],
                dtype='i4',
                attrs={"units": "m/s"}
            )
            ds.rdf["units"].predicate = "http://w3id.org/nfdi4ing/metadata4ing#hasUnit"

        res = validate_hdf(
            hdf_source=h5.hdf_filename,
            shacl_data=NUMERIC_DATASETS_SHALL_HAVE_UNIT_AND_KIND_OF_QUANTITY
        )
        self.assertFalse(res.conforms)

        with h5tbx.File() as h5:
            ds = h5.create_dataset(
                "numeric_dataset_no_unit",
                data=[1, 2, 3, 4, 5],
                dtype='i4',
                attrs={"units": "m/s", "kind_of_quantity": "Velocity"}
            )
            ds.rdf["units"].predicate = "http://w3id.org/nfdi4ing/metadata4ing#hasUnit"
            ds.rdf.object["units"] = "https://qudt.org/vocab/unit/M-PER-SEC"
            ds.rdf["kind_of_quantity"].predicate = "http://w3id.org/nfdi4ing/metadata4ing#hasKindOfQuantity"
            ds.rdf.object["kind_of_quantity"] = "http://qudt.org/vocab/quantitykind/Velocity"

        res = validate_hdf(
            hdf_source=h5.hdf_filename,
            shacl_data=NUMERIC_DATASETS_SHALL_HAVE_UNIT_AND_KIND_OF_QUANTITY
        )
        self.assertTrue(res.conforms)

    def test_hdf_file_has_creator(self):
        with h5tbx.File() as h5:
            pass

        res = validate_hdf(
            hdf_source=h5.hdf_filename,
            shacl_data=SHALL_HAVE_CREATOR
        )
        self.assertFalse(res.conforms)
        self.assertEqual(
            res.messages[0],
            "Each hdf:File must have at least one dcterms:creator which is either an IRI or a prov:Person."
        )

        # Now add the creator attribute
        with h5tbx.File(h5.hdf_filename, mode='a') as h5:
            h5.attrs["creator"] = "Matthias Probst"
            h5.frdf["creator"].predicate = DCTERMS.creator

        res = validate_hdf(
            hdf_source=h5.hdf_filename,
            shacl_data=SHALL_HAVE_CREATOR
        )
        self.assertFalse(res.conforms)
        self.assertEqual(
            res.messages[0],
            "Each dcterms:creator must be either an IRI or a prov:Person."
        )

        with h5tbx.File(h5.hdf_filename, mode='a') as h5:
            h5.attrs["creator"] = "https://orcid.org/0000-0001-8729-0482"
            h5.frdf["creator"].predicate = DCTERMS.creator
        res = validate_hdf(
            hdf_source=h5.hdf_filename,
            shacl_data=SHALL_HAVE_CREATOR
        )
        self.assertTrue(res.conforms)

        with h5tbx.File(h5.hdf_filename, mode='a') as h5:
            h5.attrs["creator"] = "Matthias Probst"
            h5.frdf["creator"].predicate = DCTERMS.creator
            h5.frdf["creator"].object = PROV.Person
        res = validate_hdf(
            hdf_source=h5.hdf_filename,
            shacl_data=SHALL_HAVE_CREATOR
        )
        self.assertTrue(res.conforms)

    def test_hdf_file_has_created_data(self):
        with h5tbx.File() as h5:
            pass

        res = validate_hdf(
            hdf_source=h5.hdf_filename,
            shacl_data=SHALL_HAVE_CREATED_DATE
        )
        self.assertFalse(res.conforms)
        self.assertEqual(
            res.messages[0],
            "Each hdf:File must have exactly one dcterms:created value of type xsd:date."
        )

        # Now add the creat attribute
        with h5tbx.File(h5.hdf_filename, mode='a') as h5:
            h5.attrs["created"] = "2025-01-10"
            h5.frdf["created"].predicate = "http://purl.org/dc/terms/created"

        res = validate_hdf(
            hdf_source=h5.hdf_filename,
            shacl_data=SHALL_HAVE_CREATED_DATE
        )
        self.assertTrue(res.conforms)

    @pytest.mark.skip(reason="Skipping for now, as pyshacl seems to have issues with complex shapes.")
    def test_shall_have_a_well_described_ssn_sensor_description(self):
        oprop = ObservableProperty(
            id="http://example.org/observable_property/1",
        )
        measurement_range1 = MeasurementRange(
            id="http://example.org/measurement_range/1",
            min_value="0",
            max_value="250",
            unit_code=Unit(
                id="http://qudt.org/vocab/unit/PA"
            )
        )
        u_pa = Unit(
            id="http://qudt.org/vocab/unit/PA"
        )
        measurement_range2 = MeasurementRange(
            id="http://example.org/measurement_range/2",
            min_value="0",
            max_value="500",
            unit_code=u_pa
        )
        accuracy_1 = Accuracy(
            id="http://example.org/accuracy/1",
            value=0.01 * 250,
            unit_code=u_pa,
            comment="Max error bound (±1%FS) for range 0–250 Pa (FS=250 Pa).@en"
        )
        accuracy_2 = Accuracy(
            id="http://example.org/accuracy/2",
            value=0.01 * 500,
            unit_code=u_pa,
            comment="Max error bound (±1%FS) for range 0–500 Pa (FS=500 Pa).@en"
        )
        capability_1 = SystemCapability(
            id="http://example.org/system_capability/1",
            hasSystemProperty=[accuracy_1, measurement_range1],
            forProperty=oprop
        )
        capability_2 = SystemCapability(
            id="http://example.org/system_capability/2",
            hasSystemProperty=[accuracy_2, measurement_range2],
            forProperty=oprop
        )
        sensor = Sensor(
            id="http://example.org/tool/KalinskyDS2-1",
            observes=oprop,
            hasSystemCapability=[capability_1, capability_2],
            label="Kalinsky Sensor TYPE DS 1@en"
        )

        ttl = sensor.serialize("ttl")
        from pyshacl import validate as pyshacl_validate
        data_graph = rdflib.Graph().parse(data=ttl, format="ttl")
        shacl_graph = rdflib.Graph().parse(
            data=SHALL_HAVE_WELL_DESCRIBED_SSN_SENSOR,
            format="ttl",
        )

        conforms, results_graph, results_text = pyshacl_validate(
            data_graph,
            shacl_graph=shacl_graph,
            inference="rdfs",
            abort_on_first=False,
            meta_shacl=False,
            advanced=False,
            debug=False
        )
        if not conforms:
            print(results_text)
        self.assertTrue(conforms)

        with h5tbx.File() as h5:
            h5.create_group("/test_group")
            ttl = sensor.serialize("ttl")
        res = validate_hdf(
            hdf_source=h5.hdf_filename,
            shacl_data=SHALL_HAVE_WELL_DESCRIBED_SSN_SENSOR
        )



from dataclasses import dataclass
from typing import Any

import rdflib


@dataclass
class DatabaseResource:
    identifier: Any
    metadata: rdflib.Graph

    def serialize(self, format, **kwargs) -> str:
        return self.metadata.serialize(format=format, **kwargs)



import pathlib
import unittest

import dotenv
import matplotlib.pyplot as plt
import rdflib
import requests
from h5rdmtoolbox.catalog import GraphDB
from ontolutils.namespacelib import QUDT_KIND, QUDT_UNIT
from ssnolib.m4i import NumericalVariable

from opencefadb import plotting
from opencefadb.core import OpenCeFaDB
from opencefadb.models.fan_curve import SemanticFanCurve

__this_dir__ = pathlib.Path(__file__).parent


class TestFanCurve(unittest.TestCase):

    def setUp(self):
        dotenv.load_dotenv(__this_dir__ / ".env", override=True)
        self.working_dir = pathlib.Path(__this_dir__ / "local-db")

    def test_get_fan_curve(self):
        try:
            graphdb = GraphDB(
                endpoint="http://localhost:7200",
                repository="OpenCeFaDB-Sandbox",
                username="admin",
                password="admin"
            )
            graphdb.get_repository_info("OpenCeFaDB-Sandbox")
        except requests.exceptions.ConnectionError as e:
            self.skipTest(f"GraphDB not available: {e}")
        res = graphdb.get_or_create_repository(__this_dir__ / "graphdb-config-sandbox.ttl")
        self.assertTrue(res)

        db = OpenCeFaDB.from_graphdb_setup(
            working_directory=self.working_dir,
            version="latest",
            sandbox=False,
            endpoint="http://localhost:7200",
            repository="OpenCeFaDB-Sandbox",
            username="admin",
            password="admin",
            add_wikidata_store=True
        )
        # db.download_metadata()
        # for file in db.rdf_directory.rglob("*.ttl"):
        #     db.rdf_store.upload_file(file)
        #     db.add_hdf_infile_index()
        # db.add_hdf_infile_index()

        # define the standard names:
        zenodo_record_ns = rdflib.namespace.Namespace("https://doi.org/10.5281/zenodo.17572275#")
        sn_mean_dp_stat = zenodo_record_ns[
            'standard_name_table/derived_standard_name/arithmetic_mean_of_difference_of_static_pressure_between_fan_outlet_and_fan_inlet']
        sn_mean_vfr = zenodo_record_ns[
            'standard_name_table/derived_standard_name/arithmetic_mean_of_fan_volume_flow_rate']
        sn_mean_nrot = zenodo_record_ns[
            'standard_name_table/derived_standard_name/arithmetic_mean_of_fan_rotational_speed']
        operating_point_standard_names = {
            sn_mean_dp_stat,
            sn_mean_vfr,
            sn_mean_nrot
        }

        # test getting fan curve data:
        n_rot = 600
        operating_point_observations = db.get_operating_point_observations(
            n_rot_speed_rpm=n_rot,
            operating_point_standard_names=operating_point_standard_names,
            standard_name_of_rotational_speed=sn_mean_nrot
        )

        fan_curve = SemanticFanCurve.from_observations(
            observations=operating_point_observations
        )
        n_scale = NumericalVariable(
            has_numerical_value=600,
            has_unit=QUDT_UNIT.REV_PER_MIN,
            has_kind_of_quantity=QUDT_KIND.RotationalVelocity
        )
        fan_curve_scaled = fan_curve.scale(
            n_scale
        )
        if n_rot == 600:
            self.assertTrue(len(fan_curve) > 40)
        elif n_rot == 1200:
            self.assertEqual(12, len(fan_curve))

        if False:
            with plotting.SingleAxis(
                    scale=1.0,
                    filename="test_fan_curve.svg",
            ) as dax:
                fan_curve.errorbar(
                    x="arithmetic_mean_of_fan_volume_flow_rate",
                    y="arithmetic_mean_of_difference_of_static_pressure_between_fan_outlet_and_fan_inlet",
                    xlabel=None,
                    ylabel=None,
                    label="Test Fan Curve",
                    marker=".",
                    linestyle='-',
                    ax=dax.ax,
                )
                fan_curve_scaled.errorbar(
                    x="arithmetic_mean_of_fan_volume_flow_rate",
                    y="arithmetic_mean_of_difference_of_static_pressure_between_fan_outlet_and_fan_inlet",
                    xlabel=None,
                    ylabel=None,
                    label="Scaled Test Fan Curve",
                    marker=".",
                    linestyle='-',
                    color="green",
                    ax=dax.ax,
                )
                plt.legend()
                plt.tight_layout()
                plt.show()



from h5rdmtoolbox import catalog as h5cat


def find_dataset_for_standard_name(
        standard_name: str,
        value_range: tuple[float, ...] = None,
) -> h5cat.SparqlQuery:
    query_str = f"""
PREFIX hdf:  <http://purl.allotrope.org/ontologies/hdf5/1.8#>
PREFIX ssno: <https://matthiasprobst.github.io/ssno#>
PREFIX xsd:  <http://www.w3.org/2001/XMLSchema#>

SELECT ?dataset ?value
WHERE {{
  BIND(<{standard_name}> AS ?standardName)

  ?dataset a hdf:Dataset ;
           ssno:hasStandardName ?standardName ;
           hdf:value ?value .
"""
    if value_range is None:
        query_str += "}\n"
    else:
        query_str += f"""
  FILTER(
    xsd:double(?value) >= {value_range[0]} &&
    xsd:double(?value) <= {value_range[1]}
  )
}}
"""
    return h5cat.SparqlQuery(
        query=query_str,
        description=f"Find datasets with {standard_name}. Returns the dataset URI and its value." if value_range is None else f"Find datasets with {standard_name} where value in {value_range}."
    )


def find_datasets_by_standard_name_and_value_range(
        standard_name: str,
        value_range: tuple[float, float] = None,
) -> h5cat.SparqlQuery:
    query_str = f"""
PREFIX hdf:  <http://purl.allotrope.org/ontologies/hdf5/1.8#>
PREFIX ssno: <https://matthiasprobst.github.io/ssno#>
PREFIX idx:  <urn:h5cat:index#>

SELECT ?hdfFile ?dataset ?value
WHERE {{
  ?dataset ssno:hasStandardName <{standard_name}> ;
           hdf:value ?value ;
           idx:inFile ?hdfFile .
"""
    if value_range is not None:
        lo, hi = value_range
        query_str += f"""
  FILTER(?value >= {lo} && ?value <= {hi})
"""
    query_str += "}\n"

    return h5cat.SparqlQuery(
        query=query_str,
        description=(
            f"Find datasets with {standard_name}."
            if value_range is None
            else f"Find datasets with {standard_name} where value in {value_range}."
        ),
    )


def find_hdf5_file_for_dataset(dataset_uri: str) -> h5cat.SparqlQuery:
    query_str = f"""
PREFIX idx: <urn:h5cat:index#>

SELECT ?hdfFile
WHERE {{
  <{dataset_uri}> idx:inFile ?hdfFile .
}}
"""
    return h5cat.SparqlQuery(
        query=query_str,
        description=f"Find HDF5 file containing dataset {dataset_uri}"
    )


def find_dataset_in_file_by_standard_name(
        hdf_file_uri: str,
        standard_name_uri: str,
) -> h5cat.SparqlQuery:
    query_str = f"""
PREFIX idx:  <urn:h5cat:index#>
PREFIX ssno: <https://matthiasprobst.github.io/ssno#>
PREFIX hdf:  <http://purl.allotrope.org/ontologies/hdf5/1.8#>
PREFIX m4i:  <http://w3id.org/nfdi4ing/metadata4ing#>
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>

SELECT ?dataset ?value ?units ?label ?altLabel ?hasSymbol ?hasMinimumValue ?hasMaximumValue
WHERE {{
  ?dataset idx:inFile <{hdf_file_uri}> ;
           hdf:value ?value ;
           ssno:hasStandardName <{standard_name_uri}> .
           
  OPTIONAL {{ ?dataset m4i:hasMinimumValue ?hasMinimumValue }}
  OPTIONAL {{ ?dataset m4i:hasMaximumValue ?hasMaximumValue }}
  OPTIONAL {{ ?dataset m4i:hasSymbol ?hasSymbol }}
  OPTIONAL {{ ?dataset m4i:hasUnit ?units }}
  OPTIONAL {{ ?dataset rdfs:label ?label }}
  OPTIONAL {{ ?dataset skos:altLabel ?altLabel }}
}}
"""
    return h5cat.SparqlQuery(
        query=query_str,
        description=f"Find datasets in {hdf_file_uri} with standard name {standard_name_uri}"
    )



UNIT_NODE_SHALL_HAVE_KIND_OF_QUANTITY = '''@prefix sh:   <http://www.w3.org/ns/shacl#> .
@prefix m4i:  <http://w3id.org/nfdi4ing/metadata4ing#> .
@prefix qudt: <http://qudt.org/schema/qudt/> .

#################################################################
# Any node that uses m4i:hasUnit must also specify kind of quantity
# and both must point to the correct QUDT classes.
#################################################################

m4i:UnitAndQuantityKindShape
  a sh:NodeShape ;
  sh:targetSubjectsOf m4i:hasUnit ;

  # m4i:hasUnit -> IRI that is a qudt:Unit
  sh:property [
    sh:path m4i:hasUnit ;
    sh:minCount 1 ;
    sh:nodeKind sh:IRI ;
    sh:class qudt:Unit ;
    sh:message "m4i:hasUnit must point to an IRI that is a qudt:Unit." ;
  ] ;

  # m4i:hasKindOfQuantity -> IRI that is a qudt:QuantityKind
  sh:property [
    sh:path m4i:hasKindOfQuantity ;
    sh:minCount 1 ;
    sh:nodeKind sh:IRI ;
    sh:class qudt:QuantityKind ;
    sh:message "If a node has m4i:hasUnit, it must also have m4i:hasKindOfQuantity pointing to an IRI that is a qudt:QuantityKind." ;
  ] .
'''


WIKIDATA_ITS_FAN_V1 = "https://www.wikidata.org/entity/Q131549102"
WIKIDATA_FAN_OPERATING_POINT = "https://www.wikidata.org/entity/Q137525225"

WIKIDATA_AFFINITY_LAWS = "https://www.wikidata.org/entity/Q632736"



"""

converts physical variables of pressure and volume flow rate to dimensionless numbers psi and phi and wise versa
from dimensionless to physical variables

"""
from typing import Union

import numpy as np
import xarray as xr


def psi(dp: Union[float, np.ndarray],
        n: Union[float, np.ndarray],
        rho: Union[float, np.ndarray],
        D: float) -> Union[float, np.ndarray]:
    """Head coefficient

    Parameters
    ----------
    dp: `float` or `np.ndarray[1d]`
        pressure head [Pa]
    n: `float` or `np.ndarray[1d]`
        revolution speed [rev/s]
    rho: `float` or `np.ndarray[1d]`
        Density of fluid in [kg/m^3]
    D: `float`
        Diameter [m] of the vane
    """
    u2 = np.pi * n * D
    return dp / (rho / 2 * u2 ** 2)


def psi2dptot(psi, n, rho, D):
    """Calculates pressure increase from pressure coefficient

    Parameters
    ----------
    psi: `float` or `np.ndarray[1d]`
        pressure coefficient [-]
    n: `float` or `np.ndarray[1d]`
        revolution speed [rev/s]
    rho: `float` or `np.ndarray[1d]`
        Density of fluid in [kg/m^3]
    D: `float`
        Diameter [m] of the vane
    """
    return psi * (0.5 * np.pi ** 2 * rho * D ** 2 * n ** 2)


def phi(vfr, n, D):
    """Flow coefficient

    Parameters
    ----------
    vfr: `float` or `np.ndarray[1d]`
        volume flow rate in [m^3/s]
    n: `float` or `np.ndarray[1d]`
        revolution speed [rev/s]
    D: `float`
        Diameter [m] of the vane
    """
    u2 = np.pi * n * D
    return vfr / (np.pi / 4 * D ** 2 * u2)


def phi2vfr(phi, n, D):
    """Calculates volume flow rate from flow coefficient

    Parameters
    ----------
    phi: `float` or `np.ndarray[1d]`
        flow coefficient [-]
    n: `float` or `np.ndarray[1d]`
        revolution speed [rev/s]
    D: `float`
        Diameter [m] of the vane
    """
    return phi * (np.pi ** 2 * D ** 3 * n) / 4


def compute_from_dataset(ds: xr.Dataset) -> xr.Dataset:
    """Compute the dimless dataset. Required data_vars: n, rho, D.
    Only phi and psi are computed, the rest is dropped"""

    if 'n' not in ds:
        raise KeyError('Missing revolution speed data_var!')

    n = ds.n.pint.quantify().pint.to('1/s').pint.dequantify()
    rho = ds.rho

    D = ds.D.pint.quantify().pint.to('m').pint.dequantify()
    _psi = psi(ds.dp_tt.pint.quantify().pint.to('Pa').pint.dequantify(),
               n=n,
               rho=rho,
               D=D)
    _phi = phi(ds.vfr.pint.quantify().pint.to('m^3/s').pint.dequantify(),
               n=n,
               D=D)
    _phi.attrs = {'long_name': 'flow coefficient',
                  'units': '',
                  'standard_name': 'fan_flow_coefficient'}
    _psi.attrs = {'long_name': 'pressure coefficient',
                  'units': '',
                  'standard_name': 'fan_pressure_coefficient'}
    dimless_ds = xr.Dataset(
        data_vars={'psi': _psi, 'phi': _phi}
    )
    dimless_ds.attrs = ds.attrs
    # dimless_ds.parent = ds
    return dimless_ds



# app.py
import base64
import json
import re
from typing import Optional

import rdflib
import streamlit as st
import streamlit.components.v1 as components
from SPARQLWrapper import SPARQLWrapper, TURTLE  # pip install SPARQLWrapper
from rdflib import Graph, URIRef, RDF
from rdflib.plugins.stores.sparqlstore import SPARQLStore

st.set_page_config(page_title="RDF Graph Explorer", layout="wide")

st.title("🔬 RDF Graph Explorer")


def find_root_nodes(graph: rdflib.Graph, *, uris_only: bool = True) -> set[URIRef]:
    """Root nodes = subjects with no incoming edges (?x ?p s)."""
    subjects = set()
    objects = set()

    for s, p, o in graph:
        subjects.add(s)
        if isinstance(o, (rdflib.term.URIRef, rdflib.term.BNode)):
            objects.add(o)

    candidates = subjects - objects
    if uris_only:
        candidates = {n for n in candidates if isinstance(n, URIRef)}
    return candidates


def get_root_node_iris(graph: rdflib.Graph) -> list[URIRef]:
    roots = find_root_nodes(graph, uris_only=True)
    return sorted(roots)


def get_root_node_iri(graph: rdflib.Graph) -> Optional[URIRef]:
    roots = get_root_node_iris(graph)
    return roots[0] if roots else None


# ---------- SPARQL / Remote loading helpers ----------

def _basic_auth_header(username: str, password: str) -> str:
    token = base64.b64encode(f"{username}:{password}".encode("utf-8")).decode("ascii")
    return f"Basic {token}"


def load_graph_from_rdflib_endpoint(endpoint_url: str, limit: int = 5000) -> Graph:
    sparql = SPARQLWrapper(endpoint_url)
    sparql.setReturnFormat(TURTLE)

    q = f"""
    CONSTRUCT {{ ?s ?p ?o }}
    WHERE      {{ ?s ?p ?o }}
    LIMIT {int(limit)}
    """
    sparql.setQuery(q)

    ttl_bytes = sparql.query().convert()  # bytes (turtle)
    if isinstance(ttl_bytes, str):
        ttl_bytes = ttl_bytes.encode("utf-8")

    g = Graph()
    g.parse(data=ttl_bytes.decode("utf-8"), format="turtle")
    return g


def load_graph_from_sparql_endpoint(
        endpoint_url: str,
        roots: Optional[list[str]] = None,
        limit: int = 5000,
        username: Optional[str] = None,
        password: Optional[str] = None,
) -> Graph:
    """
    Fetch a bounded subgraph from a SPARQL endpoint into a local Graph via CONSTRUCT.
    If roots given: fetch outgoing+incoming around roots.
    Else: fetch first `limit` triples.
    """
    headers = {}
    if username and password:
        headers["Authorization"] = _basic_auth_header(username, password)

    store = SPARQLStore(endpoint_url, headers=headers) if headers else SPARQLStore(endpoint_url)
    remote = Graph(store=store)

    local = Graph()

    if roots:
        # basic IRI wrapping for VALUES
        roots_iris = []
        for r in roots:
            r = r.strip()
            if not r:
                continue
            if r.startswith("<") and r.endswith(">"):
                roots_iris.append(r)
            else:
                roots_iris.append(f"<{r}>")

        if not roots_iris:
            return local

        values = " ".join(roots_iris)

        q = f"""
        CONSTRUCT {{
          ?root ?p ?o .
          ?s ?p2 ?root .
        }}
        WHERE {{
          VALUES ?root {{ {values} }}
          {{
            ?root ?p ?o .
          }} UNION {{
            ?s ?p2 ?root .
          }}
        }}
        LIMIT {int(limit)}
        """
    else:
        q = f"""
        CONSTRUCT {{
          ?s ?p ?o .
        }}
        WHERE {{
          ?s ?p ?o .
        }}
        LIMIT {int(limit)}
        """

    res = remote.query(q)
    if hasattr(res, "graph") and res.graph is not None:
        local += res.graph
    return local


# ---------- HTML / Vis.js ----------

GRAPH_HTML = """
<!DOCTYPE html>
<html>
<head>
  <script src="https://cdn.jsdelivr.net/npm/vis-network/standalone/umd/vis-network.min.js"></script>
  <style>
    body { margin: 0; font-family: Arial; background: #f8f9fa; }
    /* Graph area now leaves space on the right for the details panel */
    #graph { position: absolute; top: 0; left: 0; right: 320px; bottom: 0; border: none; }
    .status { position: absolute; top: 10px; left: 10px; background: white; padding: 10px;
              border-radius: 5px; box-shadow: 0 2px 10px rgba(0,0,0,0.1); font-weight: bold; z-index: 200; }
    .toolbar { position:absolute; top:10px; right:330px; z-index:210; display:flex; gap:8px; }
    .toolbar button { padding:4px 8px; font-size:11px; cursor:pointer; border-radius:4px;
                      border:1px solid #ccc; background:#fff; }
    .toolbar button:hover { background:#f0f0f0; }
    /* Node details panel on the right side */
    #payload {
      position:absolute;
      top:0;
      right:0;
      bottom:0;
      width:320px;
      overflow:auto;
      background:#fff;
      border-left:1px solid #ddd;
      padding:10px;
      z-index:100;
      opacity:0.97;
      font-size:12px;
      color:#222;
    }
    #payload h4 { margin:0 0 8px 0; font-size:13px; }
    #payload .section-title { font-weight:bold; margin-top:10px; margin-bottom:4px; }
    #payload ul { margin:4px 0 8px 16px; padding:0; }
    #payload li { margin-bottom:4px; }
    #payload code { font-size:11px; }
  </style>
</head>
<body>
  <div id="graph"></div>
  <div class="status" id="status">🖕 Double-click roots to expand</div>
  <div class="toolbar">
    <button id="cleanupBtn" title="Remove nodes without any edges">Cleanup</button>
    <button id="expandAllBtn" title="Expand all nodes in current subgraph">Expand All</button>
  </div>
  <div id="payload">
    <h4>Node details</h4>
    <div>Click a node to see its details here.</div>
  </div>

  <script>
    const nodes = new vis.DataSet({nodes_json});
    const edges = new vis.DataSet([]);
    const outgoing = {outgoing_json};
    const nodeMeta = {node_meta_json};

    let lastSelectedIri = null;

    // secondary index by local name -> list of children
    const outgoingByLocal = {};
    Object.keys(outgoing).forEach(function(k){
      try {
        const local = k.split(/[\\/\\#]/).pop();
        outgoingByLocal[local] = outgoingByLocal[local] || [];
        outgoingByLocal[local] = outgoingByLocal[local].concat(outgoing[k]);
      } catch(e) {}
    });

    const container = document.getElementById('graph');
    const data = { nodes, edges };

    const options = {
      physics: { enabled: true, solver: 'forceAtlas2Based', forceAtlas2Based: { springLength: 150 } },
      nodes: { shape: 'dot', size: 20, font: { size: 11 } },
      edges: { arrows: 'to', font: { size: 9 } }
    };

    let network;
    let expanded = new Set();

    function shortName(uri) {
      return String(uri).split('/').pop().split('#').pop();
    }

    function escapeHtml(str) {
      return String(str)
        .replace(/&/g, "&amp;")
        .replace(/</g, "&lt;")
        .replace(/>/g, "&gt;");
    }

    function showNodeDetails(id) {
      try {
        const n = nodes.get(id);
        const meta = nodeMeta[id] || {};
        const out = outgoing[id] || [];
        const payloadEl = document.getElementById('payload');
        if (!payloadEl) return;

        const label = n ? n.label : shortName(id);
        const iri = n ? n.title : String(id);
        lastSelectedIri = iri;

        const metaKeys = Object.keys(meta);
        let metaHtml = '';
        if (metaKeys.length === 0) {
          metaHtml = '<em>No literal properties available.</em>';
        } else {
          metaHtml = '<ul>' + metaKeys.map(k => {
            const values = meta[k] || [];
            const valuesText = values.map(v => escapeHtml(v)).join(', ');
            return `<li><strong>${escapeHtml(k)}</strong>: ${valuesText}</li>`;
          }).join('') + '</ul>';
        }

        let outHtml = '';
        if (!out.length) {
          outHtml = '<em>No outgoing links from this node.</em>';
        } else {
          outHtml = '<ul>' + out.map(child => {
            const pred = child.label || '';
            const targetShort = shortName(child.to);
            return `<li><strong>${escapeHtml(pred)}</strong> → ${escapeHtml(targetShort)}</li>`;
          }).join('') + '</ul>';
        }

        const html = `
          <h4>Node details</h4>
          <div style="margin-bottom:8px; display:flex; align-items:center; gap:8px;">
            <div style="flex:1 1 auto;">
              <div><strong>Label:</strong> ${escapeHtml(label)}</div>
              <div><strong>IRI:</strong>
                <code>
                  <a href="${iri}" target="_blank" rel="noopener noreferrer">
                    ${escapeHtml(iri)}
                  </a>
                </code>
              </div>
            </div>
            <button id="copyBtn" title="Copy IRI" style="padding:4px 8px; font-size:11px; cursor:pointer; border-radius:4px; border:1px solid #ccc; background:#fff;">Copy</button>
          </div>
          <div style="margin-bottom:6px;">
            <div class="section-title">Properties</div>
            ${metaHtml}
          </div>
          <div>
            <div class="section-title">Links</div>
            ${outHtml}
          </div>
        `;
        payloadEl.innerHTML = html;

        // Attach copy handler
        const copyBtn = document.getElementById('copyBtn');
        if (copyBtn) {
          copyBtn.onclick = async () => {
            const textToCopy = lastSelectedIri || iri;
            try {
              if (navigator.clipboard && navigator.clipboard.writeText) {
                await navigator.clipboard.writeText(textToCopy);
              } else {
                const ta = document.createElement('textarea');
                ta.value = textToCopy;
                ta.style.position = 'fixed';
                ta.style.top = '-1000px';
                document.body.appendChild(ta);
                ta.focus();
                ta.select();
                document.execCommand('copy');
                document.body.removeChild(ta);
              }
              const st = document.getElementById('status');
              if (st) st.innerHTML = '📋 Copied IRI to clipboard';
            } catch (e) {
              const st = document.getElementById('status');
              if (st) st.innerHTML = '⚠️ Failed to copy';
              console.log('copy error', e);
            }
          };
        }
      } catch (e) {
        console.log('showNodeDetails error', e);
      }
    }

    function expandNode(id) {
      expanded.add(id);
      const idLocal = String(id).split(/[\\/\\#]/).pop();
      const children = outgoing[id] || outgoingByLocal[idLocal] || [];
      let added = 0;

      children.forEach(child => {
        try {
          if (!nodes.get(child.to)) {
            let pos = null;
            try { pos = network.getPositions([id])[id]; } catch(e) { pos = null; }
            const nodeObj = {
              id: child.to,
              label: shortName(child.to),
              title: child.to,
              color: { background: '#3498db' }
            };
            if (pos) {
              const nx = pos.x + (Math.random() - 0.5) * 120;
              const ny = pos.y + (Math.random() - 0.5) * 120;
              nodeObj.x = nx; nodeObj.y = ny;
              nodeObj.fixed = { x: true, y: true };
              nodeObj.physics = false;
            }
            nodes.add(nodeObj);
            if (pos) {
              setTimeout(() => {
                try {
                  nodes.update({ id: child.to, fixed: { x: false, y: false }, physics: true });
                } catch(e) {}
              }, 700);
            }
          }
          const eid = id + '|' + child.to;
          try {
            if (!edges.get(eid)) edges.add({ id: eid, from: id, to: child.to, label: child.label });
          } catch(e) {
            try { edges.add({ id: eid, from: id, to: child.to, label: child.label }); } catch(e) {}
          }
          added++;
        } catch(err) {
          console.log('expandNode error', err);
        }
      });

      const st = document.getElementById('status');
      if (st) st.innerHTML = `🔓 Expanded ${shortName(id)} (+${added} children)`;
      network.fit();
      showNodeDetails(id);
    }

    function collapseNode(id) {
      expanded.delete(id);
      const children = outgoing[id] || [];
      let removed = 0;

      children.forEach(child => {
        try {
          if (nodes.get(child.to)) {
            nodes.remove(child.to);
            removed++;
          }
          const eid = id + '|' + child.to;
          if (edges.get(eid)) edges.remove(eid);
        } catch(err) {
          console.log('collapseNode error', err);
        }
      });

      const st = document.getElementById('status');
      if (st) st.innerHTML = `🔒 Collapsed ${shortName(id)} (-${removed} children)`;
      network.fit();
      showNodeDetails(id);
    }

    function cleanupIsolatedNodes() {
      try {
        const allNodeIds = nodes.getIds ? nodes.getIds() : nodes.get().map(n => n.id);
        const allEdges = edges.get();
        const connected = new Set();

        allEdges.forEach(e => {
          if (e.from != null) connected.add(e.from);
          if (e.to != null) connected.add(e.to);
        });

        const toRemove = [];
        allNodeIds.forEach(id => {
          if (!connected.has(id)) toRemove.push(id);
        });

        if (toRemove.length === 0) {
          const st = document.getElementById('status');
          if (st) st.innerHTML = '✅ Cleanup: no isolated nodes to remove';
          return;
        }

        nodes.remove(toRemove);

        const st = document.getElementById('status');
        if (st) st.innerHTML = `🧹 Cleanup: removed ${toRemove.length} isolated node(s)`;

        const payloadEl = document.getElementById('payload');
        if (payloadEl) {
          payloadEl.innerHTML = '<h4>Node details</h4><div>Click a node to see its details here.</div>';
        }
      } catch (e) {
        console.log('cleanupIsolatedNodes error', e);
      }
    }

    function expandAll() {
      try {
        const st = document.getElementById('status');
        if (st) st.innerHTML = '⏳ Expanding all nodes...';

        let prevCount = nodes.length ? nodes.length : (nodes.getIds ? nodes.getIds().length : nodes.get().length);
        let iterations = 0;
        const maxIterations = 20; // safety to avoid runaway expansion
        while (iterations < maxIterations) {
          iterations++;
          const ids = nodes.getIds ? nodes.getIds() : nodes.get().map(n => n.id);
          ids.forEach(id => {
            try { expandNode(id); } catch(e) {}
          });
          const curCount = nodes.length ? nodes.length : (nodes.getIds ? nodes.getIds().length : nodes.get().length);
          if (curCount <= prevCount) break; // no new nodes added
          prevCount = curCount;
        }
        network.fit();
        if (st) st.innerHTML = `✅ Expand All done (iterations: ${iterations}, nodes: ${prevCount})`;
      } catch (e) {
        console.log('expandAll error', e);
        const st = document.getElementById('status');
        if (st) st.innerHTML = '⚠️ Expand All failed';
      }
    }

    try {
      network = new vis.Network(container, data, options);

      // Auto-expand first root and show details
      setTimeout(() => {
        try {
          const rootIds = nodes.getIds ? nodes.getIds() : nodes.get().map(n=>n.id);
          if (rootIds && rootIds.length > 0) {
            expandNode(rootIds[0]);
            showNodeDetails(rootIds[0]);
          }
        } catch(e) { console.log('auto-expand error', e); }
      }, 400);

    } catch (e) {
      console.error('Failed to create vis.Network', e);
      const stEl = document.getElementById('status');
      if (stEl) stEl.innerText = 'Error creating network: ' + e;
    }

    // Handlers
    try {
      if (network) {
        network.on('doubleClick', ({ nodes: ids }) => {
          const id = ids && ids[0];
          if (!id) return;
          if (expanded.has(id)) collapseNode(id); else expandNode(id);
        });

        network.on('hoverNode', ({ node }) => {
          const nodeData = nodes.get(node);
          const st = document.getElementById('status');
          if (st) st.innerText = `Hover: ${nodeData ? nodeData.label : node}`;
        });

        network.on('click', ({ nodes: ids }) => {
          const id = ids && ids[0];
          if (!id) return;
          showNodeDetails(id);
        });
      }
    } catch(e) { console.log('attach handlers error', e); }

    // Toolbar
    try {
      const cleanupBtn = document.getElementById('cleanupBtn');
      if (cleanupBtn) cleanupBtn.addEventListener('click', cleanupIsolatedNodes);
      const expandAllBtn = document.getElementById('expandAllBtn');
      if (expandAllBtn) expandAllBtn.addEventListener('click', expandAll);
    } catch (e) { console.log('toolbar binding error', e); }
  </script>
</body>
</html>
"""

# ---------- Sidebar: Source selection ----------
st.sidebar.header("RDF data source")

source_mode = st.sidebar.radio(
    "Choose source",
    ["Upload file", "SPARQL endpoint (GraphDB etc.)", "Remote RDF URL"],
    index=0,
)

uploaded_file = None
sparql_endpoint_url = ""
remote_rdf_url = ""

sparql_limit = 5000
sparql_user = ""
sparql_pass = ""

if source_mode == "Upload file":
    uploaded_file = st.sidebar.file_uploader(
        "Choose an RDF file",
        type=["ttl", "nt", "rdf", "xml", "jsonld", "n3"],
    )

elif source_mode == "SPARQL endpoint (GraphDB etc.)":
    st.sidebar.caption("GraphDB example: https://localhost:7200/repositories/REPO")
    sparql_endpoint_url = st.sidebar.text_input("SPARQL endpoint URL", value="")
    sparql_limit = st.sidebar.number_input(
        "Max triples to fetch (CONSTRUCT LIMIT)",
        min_value=100,
        max_value=200000,
        value=5000,
        step=100,
    )
    with st.sidebar.expander("Auth (optional)"):
        sparql_user = st.text_input("Username", value="")
        sparql_pass = st.text_input("Password", value="", type="password")

elif source_mode == "Remote RDF URL":
    st.sidebar.caption("Any RDF document URL rdflib can parse (TTL/RDFXML/JSON-LD/etc.)")
    remote_rdf_url = st.sidebar.text_input("RDF document URL", value="")
    sparql_limit = st.sidebar.number_input(
        "Max triples to fetch (CONSTRUCT LIMIT)",
        min_value=100,
        max_value=200000,
        value=5000,
        step=100,
    )

# Root selection (shared for all modes; especially important for SPARQL)
st.sidebar.header("Root IRI (optional)")

# Initialize session state for root IRIs list
if 'root_iris' not in st.session_state:
    st.session_state['root_iris'] = []

# Helper to normalize/prefix display using a temporary Graph namespace manager
def iri_to_prefixed(iri: str) -> str:
    try:
        tmpg = Graph()
        # reuse some common prefixes if available in loaded data later; for now rdflib defaults
        q = tmpg.namespace_manager.compute_qname(URIRef(iri))
        if q and len(q) == 3:
            prefix, namespace, lname = q
            return f"{prefix}:{lname}"
    except Exception:
        pass
    # fallback to local name
    try:
        return str(iri).split('/')[-1].split('#')[-1]
    except Exception:
        return iri

# Input field for a single IRI paste; add on Enter
def _on_add_iri():
    v = st.session_state.get('root_iri_input', '').strip()
    if v:
        # strip optional angle brackets
        if v.startswith('<') and v.endswith('>'):
            v = v[1:-1]
        # avoid duplicates
        if v not in st.session_state['root_iris']:
            st.session_state['root_iris'].append(v)
        st.session_state['root_iri_input'] = ''

st.sidebar.text_input(
    "Paste a Root IRI",
    key='root_iri_input',
    placeholder='https://example.org/resource#Thing',
    on_change=_on_add_iri
)
col_add, col_clear = st.sidebar.columns([1,1])
with col_add:
    if st.button("Add"):
        _on_add_iri()
with col_clear:
    if st.button("Clear"):
        st.session_state['root_iris'] = []

# Render current list with prefixed view and remove buttons
if st.session_state['root_iris']:
    st.sidebar.markdown("Current Roots:")
    for idx, iri in enumerate(st.session_state['root_iris']):
        pref = iri_to_prefixed(iri)
        # inline text + remove button, same line
        c_text, c_btn = st.sidebar.columns([8,1])
        with c_text:
            st.markdown(f"- `{pref}`", help=iri)
        with c_btn:
            if st.button("✕", key=f"rm_{idx}", help="Remove"):
                st.session_state['root_iris'].pop(idx)
                st.rerun()
else:
    st.sidebar.caption("No roots added yet. Paste an IRI above and press Enter or Add.")

# Use session roots downstream
root_iris_from_input: list[str] = st.session_state['root_iris'][:]
# ---------- Main: Graph Visualization ----------

g: Optional[Graph] = None
ttl_content: Optional[str] = None  # only for debug display when uploading

if source_mode == "Upload file":
    if uploaded_file is None:
        st.info("Upload an RDF file to visualize it.")
        st.stop()

    ttl_bytes = uploaded_file.read()
    try:
        ttl_content = ttl_bytes.decode("utf-8")
    except Exception:
        try:
            ttl_content = ttl_bytes.decode("latin-1")
        except Exception:
            ttl_content = ttl_bytes.decode("utf-8", errors="ignore")

    with st.spinner("🔄 Parsing uploaded RDF with RDFLib..."):
        g = Graph()
        try:
            g.parse(data=ttl_content, format="turtle")
        except Exception:
            # let rdflib guess
            g.parse(data=ttl_content)

elif source_mode == "Remote RDF URL":
    if not remote_rdf_url.strip():
        st.info("Enter an RDF document URL to visualize it.")
        st.stop()
    with st.spinner("🌐 Fetching + parsing remote RDF URL..."):
        g = load_graph_from_rdflib_endpoint(remote_rdf_url.strip(), limit=sparql_limit)

elif source_mode == "SPARQL endpoint (GraphDB etc.)":
    if not sparql_endpoint_url.strip():
        st.info("Enter a SPARQL endpoint URL to visualize it.")
        st.stop()
    with st.spinner("🔌 Fetching subgraph from SPARQL endpoint (CONSTRUCT)..."):
        g = load_graph_from_sparql_endpoint(
            endpoint_url=sparql_endpoint_url.strip(),
            roots=root_iris_from_input if root_iris_from_input else None,
            limit=int(sparql_limit),
            username=sparql_user.strip() or None,
            password=sparql_pass or None,
        )

if g is None:
    st.error("Graph could not be loaded.")
    st.stop()

if len(g) == 0:
    st.warning("Loaded graph has 0 triples. Try increasing the LIMIT or providing Root IRIs (for SPARQL).")
    st.stop()

# helper: local name using namespace manager
def local_name(uri, graph):
    try:
        prefix, namespace, lname = graph.namespace_manager.compute_qname(URIRef(uri))
        return lname
    except Exception:
        return str(uri).split("/")[-1].split("#")[-1]

# Build nodes and outgoing mapping
all_nodes: dict[str, dict] = {}
outgoing_triples: dict[str, list[dict]] = {}
incoming_triples: dict[str, list[dict]] = {}

for s, p, o in g:
    s_str, p_str, o_str = str(s), str(p), str(o)

    if s_str not in all_nodes:
        label = local_name(s_str, g)
        is_class = bool(list(g.objects(URIRef(s_str), RDF.type)))
        all_nodes[s_str] = {
            "id": s_str,
            "label": label[:40],
            "title": s_str,
            "color": "#e74c3c" if is_class else "#3498db",
            "meta": {},
        }

    if isinstance(o, URIRef):
        if o_str not in all_nodes:
            o_label = local_name(o_str, g)
            all_nodes[o_str] = {
                "id": o_str,
                "label": o_label[:40],
                "title": o_str,
                "color": "#3498db",
                "meta": {},
            }

        edge_label = local_name(p_str, g)[:24]
        outgoing_triples.setdefault(s_str, []).append({"to": o_str, "label": edge_label})

        incoming_triples.setdefault(o_str, []).append({"from": s_str, "label": edge_label})

        try:
            if URIRef(p_str) == RDF.type:
                outgoing_triples.setdefault(o_str, []).append({"to": s_str, "label": "instance"});
        except Exception:
            pass

    else:
        try:
            pred_name = local_name(p_str, g)[:40]
            meta = all_nodes[s_str].setdefault("meta", {});
            meta.setdefault(pred_name, []).append(o_str);
        except Exception:
            pass

# Determine initial nodes shown:
# - If user entered root IRIs: show those that exist in loaded graph.
# - Else: show up to 20 nodes.
roots: list[str] = []
if root_iris_from_input:
    matched = [r for r in root_iris_from_input if r in all_nodes]
    if matched:
        roots = matched
    else:
        st.warning("None of the provided Root IRIs were found in the loaded subgraph. Showing first nodes instead.")
        roots = list(all_nodes.keys())[:20]
else:
    roots = list(all_nodes.keys())[:20]

initial_nodes: list[dict] = []
for root in roots[:12]:
    initial_nodes.append(all_nodes[root].copy())

nodes_json = json.dumps(initial_nodes)
outgoing_json = json.dumps(outgoing_triples)
node_meta = {k: v.get("meta", {}) for k, v in all_nodes.items()}
node_meta_json = json.dumps(node_meta)

if not initial_nodes:
    st.warning("No nodes found — check parsing / endpoint query.");
    if ttl_content:
        st.markdown("**TTL (first 2000 chars)**");
        st.code(ttl_content[:2000]);
    st.stop()

# Render Vis.js graph
html = GRAPH_HTML.replace("{nodes_json}", nodes_json)
html = html.replace("{outgoing_json}", outgoing_json)
html = html.replace("{node_meta_json}", node_meta_json)

# NOTE: If you couldn't see the plot before, it was often because
# components.html wasn't called (or g was empty / st.stop before render).
components.html(html, height=750, scrolling=False)

# Summary
st.subheader("Graph summary")
st.write(f"Triples: {len(g)} — Nodes: {len(all_nodes)} — Roots shown: {len(initial_nodes)} — Outgoing entries: {len(outgoing_triples)}")

st.markdown("---")
st.caption("🐍 RDFLib + Vis.js | `pip install streamlit rdflib` | 🔬 Hierarchical RDF exploration")



"""Open Centrifugal Fan Database.

OpenCeFaDB is a database for centrifugal fan data. It is designed to store and retrieve data from different sources
and provide a common interface to access the data. The database is based on the Generic Linked Database (GLDB)
framework and uses a combination of SQL or NoSQL databases to store large amounts of raw data and RDF databases
to store the metadata.

Users can choose which specific database to use. There are some predefined configurations available, but users can
define their own database interfaces, leaving room for integration with their own workflows and databases.
"""
import json
import logging
import pathlib
from logging.handlers import RotatingFileHandler
from typing import Union, Dict

import appdirs

from . import models

__this_dir__ = pathlib.Path(__file__).parent

from ._version import __version__

USER_LOG_DIR = pathlib.Path(appdirs.user_log_dir('opencefadb', version=__version__))
USER_LOG_DIR.mkdir(parents=True, exist_ok=True)

DEFAULT_LOGGING_LEVEL = logging.INFO
_formatter = logging.Formatter(
    '%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d_%H:%M:%S')

_stream_handler = logging.StreamHandler()
_stream_handler.setLevel(DEFAULT_LOGGING_LEVEL)
_stream_handler.setFormatter(_formatter)

_file_handler = RotatingFileHandler(USER_LOG_DIR / 'opencefadb.log')
_file_handler.setLevel(logging.DEBUG)  # log everything to file!
_file_handler.setFormatter(_formatter)

logger = logging.getLogger("opencefadb")
logger.setLevel(DEFAULT_LOGGING_LEVEL)
logger.addHandler(_stream_handler)
logger.addHandler(_file_handler)


def set_logging_level(level: Union[int, str]):
    """Set the log level."""
    _logger = logging.getLogger("opencefadb")
    _logger.setLevel(level)
    for handler in _logger.handlers:
        handler.setLevel(DEFAULT_LOGGING_LEVEL)

    return _logger.level


GRAPH_DB_CONFIG_FILENAME = __this_dir__ / "../data/graphdb-config.ttl"

from .core import OpenCeFaDB

from .utils import opencefa_print


def get_package_meta() -> Dict:
    """Reads codemeta.json and returns it as dict"""
    with open(__this_dir__ / '../codemeta.json', 'r') as f:
        codemeta = json.loads(f.read())
    return codemeta


__all__ = ["set_logging_level", "OpenCeFaDB", "models", "opencefa_print"]



import pathlib
from typing import Union

import h5py
from h5rdmtoolbox.ld.shacl import validate_hdf


def validate_measurement_hdf(source: Union[pathlib.Path, str, h5py.File]):
    from .shacl import MEASUREMENT_HDF_FILE_CONTENT_SHACL_DEFINITIONS
    for _shacl in MEASUREMENT_HDF_FILE_CONTENT_SHACL_DEFINITIONS:
        validate_hdf(hdf_source=source, shacl_data=_shacl)



import json.decoder
import logging
import pathlib

import rdflib
from rdflib import Graph
from rdflib.plugins.stores.sparqlstore import SPARQLStore

from .request_utils import _post_request

logger = logging.getLogger("pygraphdb")

CONTENT_TYPE = {
    ".jsonld": "application/ld+json",
    ".ttl": "text/turtle",
}


class GraphDBRepository:

    def __init__(self, params):
        self._params = params
        store = SPARQLStore(auth=self._params["auth"])
        store.open(self._params["uri"])
        self.graph = Graph(store=store, identifier=None, bind_namespaces="none")

    def __getitem__(self, item):
        return self._params[item]

    def upload_file(
            self,
            filename: pathlib.Path,
            check_for_blank_nodes=True,
            use_base_iri="https://local.org/"
    ):
        logger.debug(f"Uploading file {filename} to {self['id']} ...")
        filename = pathlib.Path(filename).resolve().absolute()
        if not filename.exists():
            raise FileNotFoundError(f"File '{filename}' does not exist")

        if check_for_blank_nodes:
            has_blank_nodes, (s, p, o) = _check_for_blank_nodes(filename)
            if has_blank_nodes and use_base_iri:
                with open(str(filename), 'r', encoding='utf-8') as f:
                    content = f.read()
                content = content.replace("_:b", f"{use_base_iri}b")
                with open(str(filename), 'w', encoding='utf-8') as f:
                    f.write(content)
            elif has_blank_nodes and not use_base_iri:
                raise ValueError(f"File '{filename}' contains blank nodes. Blank nodes are not supported: "
                                 f"{s}, {p}, {o}")

        headers = {"Content-Type": CONTENT_TYPE[filename.suffix]}

        repo_id = self['id']

        with open(filename, 'rb') as f:
            response = _post_request(self["uri"] + "/statements", headers=headers, data=f, auth=self["auth"])
        if response.status_code == 204:
            logger.info(f"File '{filename}' successfully uploaded to GraphDB repository '{repo_id}'")
        else:
            logger.error(f"Could not upload file {filename}: {response.status_code} - {response.text}")
        logger.debug(f"Done uploading file {filename} to {repo_id} ...")
        return response


def _check_for_blank_nodes(filename):
    logger.debug(f"Checking file {filename} for blank nodes ...")
    g = rdflib.Graph()
    try:
        g.parse(source=filename)
    except json.decoder.JSONDecodeError as e:
        logger.error(f"Error parsing file {filename}: {e}")
        raise e
    for s, p, o in g.triples((None, None, None)):
        if isinstance(s, rdflib.BNode) or isinstance(p, rdflib.BNode) or isinstance(o, rdflib.BNode):
            return True, (s, p, o)
    return False, (None, None, None)



from ontolutils.ex.ssn import Observation

from ontolutils.namespacelib import QUDT_KIND

EXPECTED_PRESSURE_QUANTITY_KINDS = [
    QUDT_KIND.Pressure,
    QUDT_KIND.StaticPressure,
    QUDT_KIND.TotalPressure,
]


def is_operating_point(observation: Observation, verbose: bool = False) -> bool:
    """Check if an observation is a fan operating point.

    Args:
        observation (Observation): The observation to check.
        the observation must have minimum three numerical variables with the following
        quantity kinds:
            - air_flow_rate
            - static_pressure_difference
            - rotational_speed
        verbose (bool): If True, print missing quantity kinds.
    Returns:
        bool: True if the observation is a fan operating point, False otherwise.
    """
    has_results = observation.has_result
    if not has_results or len(has_results) < 3:
        return False
    has_vfr_kind = False
    has_dp_kind = False
    has_n_kind = False
    for result in has_results:
        num_var = result.has_numerical_variable
        if num_var is None:
            continue
        if num_var.is_kind_of_quantity(QUDT_KIND.VolumeFlowRate):
            has_vfr_kind = True
            continue
        if num_var.is_kind_of_quantity(QUDT_KIND.RotationalVelocity):
            has_n_kind = True
            continue
        for EPQK in EXPECTED_PRESSURE_QUANTITY_KINDS:
            if num_var.is_kind_of_quantity(EPQK):
                has_dp_kind = True
                break
    if verbose:
        if not has_vfr_kind:
            print("Missing Volume Flow Rate kind")
        if not has_dp_kind:
            print("Missing Pressure kind")
        if not has_n_kind:
            print("Missing Rotational Velocity kind")
    return has_vfr_kind and has_dp_kind and has_n_kind



from typing import Union

import rdflib
from h5rdmtoolbox import catalog as h5cat
from pydantic import HttpUrl

from . import hdf, fan

__all__ = [
    "hdf",
    "fan",
]

SELECT_ALL = h5cat.SparqlQuery("SELECT * WHERE {?s ?p ?o}", description="Selects all triples in the RDF database")


def get_m4i_parameters(subject_uri: Union[str, HttpUrl, rdflib.URIRef], limit: int = None):
    """Returns a SPARQL query that selects all M4I parameters of the given subject URI."""
    subject_uri = str(HttpUrl(subject_uri))
    query_str = f"""
    PREFIX ssno: <https://matthiasprobst.github.io/ssno#>
    PREFIX m4i: <https://w3id.org/m4i/ns#>

    SELECT ?parameter ?property ?value
    WHERE {{
        <{subject_uri}> a ?type .
        <{subject_uri}> m4i:hasParameter ?parameter .
        ?parameter ?property ?value .
    }}
    ORDER BY ?parameter
    """
    if limit is not None:
        query_str += f"LIMIT {limit}\n"

    return h5cat.SparqlQuery(
        query=query_str,
        description=f"Selects all M4I parameters of the target URI {subject_uri}"
    )

def get_properties(
        subject_uri: Union[str, HttpUrl, rdflib.URIRef],
        cls_uri: Union[str, HttpUrl, rdflib.URIRef] = None,
        limit: int = None
):
    """Returns a SPARQL query that selects all properties of the given subject URI. If cls_uri is provided,
    only properties for which the subject is of the given class are returned."""
    subject_uri = str(HttpUrl(subject_uri))
    if cls_uri is None:
        query_str = f"""
    PREFIX ssno: <https://matthiasprobst.github.io/ssno#>

    SELECT ?property ?value
    WHERE {{
        <{subject_uri}> ?property ?value .
    }}
    ORDER BY ?property
    """
    else:
        query_str = f"""
    SELECT ?property ?value
    WHERE {{
        <{subject_uri}> a {cls_uri} .
        <{subject_uri}> ?property ?value .
    }}
    ORDER BY ?property
    """
    if limit is not None:
        query_str += f"LIMIT {limit}\n"

    return h5cat.SparqlQuery(
        query=query_str,
        description=f"Selects all properties of the target URI {subject_uri}"
    )



import pathlib
import unittest

import dotenv
import rdflib
import requests
from h5rdmtoolbox import catalog as h5cat
from h5rdmtoolbox.catalog import HDF5SqlDB, GraphDB
from h5rdmtoolbox.catalog.profiles import IS_VALID_CATALOG_SHACL

from opencefadb import OpenCeFaDB
from opencefadb.sparql_templates.fan import SELECT_FAN_PROPERTIES
from opencefadb.stores import RDFFileStore

__this_dir__ = pathlib.Path(__file__).parent


class TestInitDatabase(unittest.TestCase):

    def setUp(self):
        dotenv.load_dotenv(__this_dir__ / ".env", override=True)
        self.working_dir = pathlib.Path(__this_dir__ / "local-db")

    def test_database_with_graphdb(self):
        try:
            graphdb = GraphDB(
                endpoint="http://localhost:7200",
                repository="OpenCeFaDB-Sandbox",
                username="admin",
                password="admin"
            )
            graphdb.get_repository_info("OpenCeFaDB-Sandbox")
        except requests.exceptions.ConnectionError as e:
            self.skipTest(f"GraphDB not available: {e}")

        db = OpenCeFaDB.from_graphdb_setup(
            working_directory=self.working_dir,
            version="latest",
            sandbox=False,
            endpoint="http://localhost:7200",
            repository="OpenCeFaDB-Sandbox",
            username="admin",
            password="admin",
            add_wikidata_store=True
        )
        self.assertEqual(db.catalog.version, "1.0.0")
        res = h5cat.RemoteSparqlQuery(
            "SELECT * WHERE { ?s ?p ?o }",
            description="Selects all triples in the RDF database"
        ).execute(db.main_rdf_store)

        count = len(res.data)
        tripels = db.main_rdf_store.count_triples(key="total")
        self.assertEqual(count, tripels)

        res = h5cat.RemoteSparqlQuery(
            SELECT_FAN_PROPERTIES.query,
            description="Selects all properties of the fan"
        ).execute(db.rdf_store)
        self.assertEqual(88, len(res.data))

        # find all names of persons in the database
        res = h5cat.RemoteSparqlQuery(
            """
            PREFIX foaf: <http://xmlns.com/foaf/0.1/>
            PREFIX prov: <http://www.w3.org/ns/prov#>
    
            SELECT ?s
            WHERE {
                ?s a prov:Person .
            }
            """,
            description="Selects names of all persons in the database"
        ).execute(db.main_rdf_store)
        self.assertEqual(2, len(res.data))

        # find a hdf5 dataset based on a hdf attribute standard_name="volume_flow_rate":
        res = h5cat.SparqlQuery(
            """
            PREFIX h5py: <http://purl.org/nfdi4ing/h5py#>
            PREFIX dcterms: <http://purl.org/dc/terms/>
            
            SELECT ?dataset ?title
            WHERE {
                ?dataset a h5py:Dataset .
                ?dataset dcterms:title ?title .
                ?dataset h5py:standard_name "volume_flow_rate" .
            }
            """,
            description="Finds a hdf5 dataset based on a hdf attribute standard_name='volume_flow_rate'"
        ).execute(db.rdf_store)

    def test_database_with_rdflib_store(self):
        db = OpenCeFaDB(working_directory=self.working_dir, version="latest", sandbox=False)
        self.assertEqual(db.catalog.version, "1.0.0")

        metadata_store = RDFFileStore(
            data_dir=self.working_dir / "metadata",
            recursive_exploration=True,
            formats="ttl"
        )

        db.add_main_rdf_store(metadata_store)
        db.download_metadata()
        metadata_store.populate()
        db.add_wikidata_store(augment_main_rdf_store=True)
        res = SELECT_FAN_PROPERTIES.execute(db.main_rdf_store)
        self.assertEqual(88, len(res.data))

        query_d1_value = """PREFIX m4i: <http://w3id.org/nfdi4ing/metadata4ing#>
        SELECT ?value
        WHERE {
            <https://doi.org/10.5281/zenodo.17871736#D1> m4i:hasNumericalValue ?value  .
        }
        """
        new_sparql = h5cat.SparqlQuery(
            query_d1_value,
            description="Selects all properties of a specific dataset"
        )
        res = new_sparql.execute(db.main_rdf_store)
        self.assertTrue(138.0, res.data["value"][0])

        find_first_names = """
                PREFIX foaf: <http://xmlns.com/foaf/0.1/>
                SELECT ?firstName
                WHERE {
                    ?s a prov:Person .
                    ?s foaf:firstName ?firstName .
                }
                """
        bindings = db.main_rdf_store.graph.query(find_first_names).bindings
        names = [str(row[rdflib.Variable("firstName")]) for row in bindings]
        self.assertIn("Matthias", names)
        self.assertIn("Balazs", names)

        find_snt_title = """
                PREFIX ex: <https://doi.org/10.5281/zenodo.17271932#>
                PREFIX ssno: <https://matthiasprobst.github.io/ssno#>
                PREFIX dcterms: <http://purl.org/dc/terms/>
    
                SELECT ?title
                WHERE {
                    ?s a ssno:StandardNameTable .
                    ?s dcterms:title ?title .
                }
                """
        bindings = db.main_rdf_store.graph.query(find_snt_title).bindings
        titles = [str(row[rdflib.Variable("title")]) for row in bindings]
        self.assertEqual(
            titles[0],
            "Standard Name Table for the Property Descriptions of Centrifugal Fans"
        )

    def test_config_validation(self):
        catalog = OpenCeFaDB.download_catalog(
            version="latest",
            target_directory=self.working_dir,
            sandbox=False,
            validate=False
        )
        validation_result = catalog.validate(shacl_data=IS_VALID_CATALOG_SHACL)
        if not validation_result.conforms:
            print("SHACL validation results:")
            print(validation_result.results_text)
        self.assertTrue(validation_result)




SHALL_HAVE_WELL_DESCRIBED_SSN_SENSOR_WITH_SYSTEM_CAPABILITY = '''@prefix ex: <https://example.org/> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix schema: <http://schema.org/> .
@prefix sosa: <http://www.w3.org/ns/sosa/> .
@prefix ssn: <http://www.w3.org/ns/ssn/> .
@prefix ssn_system: <http://www.w3.org/ns/ssn/systems/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

#################################################################
# Top-level sensor shape
#################################################################

ex:SensorMetadataShape
  a sh:NodeShape ;
  sh:targetClass sosa:Sensor ;

  sh:property [
    sh:path rdfs:label ;
    sh:minCount 1 ;
    sh:nodeKind sh:Literal ;
    sh:message "Sensor must have rdfs:label as a literal." ;
  ] ;

  sh:property [
    sh:path sosa:observes ;
    sh:minCount 1 ;
    sh:class sosa:ObservableProperty ;
    sh:message "Sensor must sosa:observes at least one sosa:ObservableProperty." ;
  ] ;

  sh:property [
    sh:path ssn_system:hasSystemCapability ;
    sh:minCount 1 ;
    sh:node ex:SystemCapabilityShape ;
    sh:message "Sensor must have at least one ssn_system:SystemCapability." ;
  ] .

#################################################################
# Capability shape (one range, optional accuracy)
#################################################################

ex:SystemCapabilityShape
  a sh:NodeShape ;
  sh:class ssn_system:SystemCapability ;

  sh:property [
    sh:path ssn:forProperty ;
    sh:minCount 1 ;
    sh:maxCount 1 ;
    sh:class sosa:ObservableProperty ;
    sh:message "Each capability must have exactly one ssn:forProperty (an ObservableProperty)." ;
  ] ;

  # Exactly one MeasurementRange per capability
  sh:property [
    sh:path ssn_system:hasSystemProperty ;
    sh:qualifiedValueShape ex:MeasurementRangeShape ;
    sh:qualifiedMinCount 1 ;
    sh:qualifiedMaxCount 1 ;
    sh:message "Each SystemCapability must have exactly one ssn_system:MeasurementRange as a system property." ;
  ] ;

  # Optional Accuracy per capability (0..1), numeric or textual spec allowed
  sh:property [
    sh:path ssn_system:hasSystemProperty ;
    sh:qualifiedValueShape ex:AccuracyShape ;
    sh:qualifiedMinCount 0 ;
    sh:qualifiedMaxCount 1 ;
    sh:message "Each SystemCapability may have at most one ssn_system:Accuracy as a system property." ;
  ] ;

  # Ensure forProperty is among the sensor's observed properties
  sh:sparql [
    a sh:SPARQLConstraint ;
    sh:message "Capability ssn:forProperty must be one of the parent sensor's sosa:observes properties." ;
    sh:select """
      SELECT $this
      WHERE {
        ?sensor ssn_system:hasSystemCapability $this ;
                sosa:observes ?obs .
        $this ssn:forProperty ?p .
        FILTER NOT EXISTS { ?sensor sosa:observes ?p }
      }
    """ ;
  ] .

#################################################################
# MeasurementRange shape
#################################################################

ex:MeasurementRangeShape
  a sh:NodeShape ;
  sh:class ssn_system:MeasurementRange ;

  sh:property [
    sh:path schema:minValue ;
    sh:minCount 1 ;
    sh:maxCount 1 ;
    sh:datatype xsd:double ;
    sh:message "MeasurementRange must have exactly one schema:minValue (xsd:double)." ;
  ] ;

  sh:property [
    sh:path schema:maxValue ;
    sh:minCount 1 ;
    sh:maxCount 1 ;
    sh:datatype xsd:double ;
    sh:message "MeasurementRange must have exactly one schema:maxValue (xsd:double)." ;
  ] ;

  sh:property [
    sh:path schema:unitCode ;
    sh:minCount 1 ;
    sh:maxCount 1 ;
    sh:nodeKind sh:IRI ;
    sh:message "MeasurementRange must have exactly one schema:unitCode (IRI)." ;
  ] ;

  # Sanity check: min < max
  sh:sparql [
    a sh:SPARQLConstraint ;
    sh:message "MeasurementRange must satisfy minValue < maxValue." ;
    sh:select """
      SELECT $this
      WHERE {
        $this schema:minValue ?min ;
              schema:maxValue ?max .
        FILTER (?min >= ?max)
      }
    """ ;
  ] .

#################################################################
# Accuracy shape (supports numeric OR textual spec)
#################################################################

ex:AccuracyShape
  a sh:NodeShape ;
  sh:class ssn_system:Accuracy ;

  # Either numeric accuracy...
  sh:or (
    [
      sh:property [
        sh:path schema:value ;
        sh:minCount 1 ;
        sh:maxCount 1 ;
        sh:datatype xsd:double ;
        sh:message "If using numeric accuracy, provide schema:value as xsd:double." ;
      ] ;
      sh:property [
        sh:path schema:unitCode ;
        sh:minCount 1 ;
        sh:maxCount 1 ;
        sh:nodeKind sh:IRI ;
        sh:message "If using numeric accuracy, provide schema:unitCode as IRI." ;
      ] ;
    ]
    # ...or textual spec (good for PT100 class formulas, etc.)
    [
      sh:property [
        sh:path rdfs:comment ;
        sh:minCount 1 ;
        sh:datatype rdf:langString ;
        sh:message "If using textual accuracy spec, provide rdfs:comment with a language tag." ;
      ] ;
    ]
    [
      sh:property [
        sh:path schema:description ;
        sh:minCount 1 ;
        sh:datatype rdf:langString ;
        sh:message "If using textual accuracy spec, provide schema:description with a language tag." ;
      ] ;
    ]
  ) .

'''

SHALL_HAVE_WELL_DESCRIBED_SSN_SENSOR = '''@prefix ex: <https://example.org/> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix schema: <http://schema.org/> .
@prefix sosa: <http://www.w3.org/ns/sosa/> .
@prefix ssn: <http://www.w3.org/ns/ssn/> .
@prefix ssn_system: <http://www.w3.org/ns/ssn/systems/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

#################################################################
# Top-level sensor shape
#################################################################

ex:SensorMetadataShape
  a sh:NodeShape ;
  sh:targetClass sosa:Sensor ;

  sh:property [
    sh:path rdfs:label ;
    sh:minCount 1 ;
    sh:nodeKind sh:Literal ;
    sh:message "Sensor must have rdfs:label as a literal." ;
  ] ;

  sh:property [
    sh:path sosa:observes ;
    sh:minCount 1 ;
    sh:class sosa:ObservableProperty ;
    sh:message "Sensor must sosa:observes at least one sosa:ObservableProperty." ;
  ] .

'''
